import email.utils
import os
import re
import types
import urllib.parse
from collections.abc import Iterator
from contextlib import suppress
from functools import cache
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Union,
    cast,
)

import fmf.context
from click import echo

import tmt.export
import tmt.identifier
import tmt.utils
import tmt.utils.git
from tmt.utils import ConvertError, Path
from tmt.utils.structured_field import StructuredField
from tmt.utils.themes import style

if TYPE_CHECKING:
    import tmt.base


NITRATE_TRACKER_ID = 69  # ID of nitrate in RH's bugzilla
DEFAULT_NITRATE_CATEGORY = 'Sanity'

WARNING = """
Test case has been migrated to git. Any changes made here might be overwritten.
See: https://tmt.readthedocs.io/en/latest/questions.html#nitrate-migration
""".lstrip()

gssapi: Optional[types.ModuleType] = None
nitrate: Optional[types.ModuleType] = None

# FIXME: Any - https://github.com/teemtee/tmt/issues/1602

# Until nitrate gets its own annotations and recognizable imports...
Nitrate = Any
NitrateTestPlan = Any
NitrateTestCase = Any

DEFAULT_PRODUCT: Any = None

SectionsReturnType = tuple[str, str, str, str]
HeadingsType = list[list[Union[int, str]]]
SectionsHeadingsType = dict[str, HeadingsType]

# TODO: why this exists?
log = fmf.utils.Logging('tmt').logger


def import_nitrate() -> Nitrate:
    """
    Conditionally import the nitrate module
    """

    # Need to import nitrate only when really needed. Otherwise we get
    # traceback when nitrate not installed or config file not available.
    # And we want to keep the core tmt package with minimal dependencies.
    try:
        global nitrate, DEFAULT_PRODUCT, gssapi
        import gssapi
        import nitrate

        assert nitrate
        DEFAULT_PRODUCT = nitrate.Product(name='RHEL Tests')
        return nitrate
    except ImportError:
        raise ConvertError("Install tmt+test-convert to export tests to nitrate.")
    # FIXME: ignore[union-attr]: https://github.com/teemtee/tmt/issues/1616
    except nitrate.NitrateError as error:  # type: ignore[union-attr]
        raise ConvertError(error)


def _nitrate_find_fmf_testcases(test: 'tmt.Test') -> Iterator[Any]:
    """
    Find all Nitrate test cases with the same fmf identifier

    All component general plans are explored for possible duplicates.
    """

    import tmt.base

    assert nitrate
    for component in test.component:
        try:
            for testcase in find_general_plan(component).testcases:
                struct_field = StructuredField(testcase.notes)
                try:
                    fmf_id = tmt.base.FmfId.from_spec(
                        cast(tmt.base._RawFmfId, tmt.utils.yaml_to_dict(struct_field.get('fmf')))
                    )
                    if fmf_id == test.fmf_id:
                        echo(
                            style(
                                f"Existing test case '{testcase.identifier}' "
                                f"found for given fmf id.",
                                fg='magenta',
                            )
                        )
                        yield testcase
                except tmt.utils.StructuredFieldError:
                    pass
        except nitrate.NitrateError:
            pass


def convert_manual_to_nitrate(test_md: Path) -> SectionsReturnType:
    """
    Convert Markdown document to html sections.

    These sections can be exported to nitrate.
    Expects: Markdown document as a file.
    Returns: tuple of (step, expect, setup, cleanup) sections
    as html strings.
    """

    import tmt.base

    sections_headings: SectionsHeadingsType = {
        heading: []
        for heading_list in tmt.base.SECTIONS_HEADINGS.values()
        for heading in heading_list
    }

    html = tmt.utils.markdown_to_html(test_md)
    html_splitlines = html.splitlines()

    for key in sections_headings:
        result: HeadingsType = []
        i = 0
        while html_splitlines:
            try:
                if re.search("^" + key + "$", html_splitlines[i]):
                    html_content = ''
                    if key.startswith('<h1>Test'):
                        html_content = (
                            html_splitlines[i].replace('<h1>', '<b>').replace('</h1>', '</b>')
                        )
                    for j in range(i + 1, len(html_splitlines)):
                        if re.search("^<h[1-4]>(.+?)</h[1-4]>$", html_splitlines[j]):
                            result.append([i, html_content])
                            i = j - 1
                            break
                        html_content += html_splitlines[j] + "\n"
                        # Check end of the file
                        if j + 1 == len(html_splitlines):
                            result.append([i, html_content])
            except IndexError:
                sections_headings[key] = result
                break
            i += 1
            if i >= len(html_splitlines):
                sections_headings[key] = result
                break

    def concatenate_headings_content(headings: tuple[str, ...]) -> HeadingsType:
        content = []
        for v in headings:
            content += sections_headings[v]
        return content

    def enumerate_content(content: HeadingsType) -> HeadingsType:
        # for sorting convert the index to integer, but keep whole list as list of strings
        content.sort(key=lambda a: int(a[0]))
        for i, entry in enumerate(content):
            entry[1] = f"<p>Step {i + 1}.</p>" + str(entry[1])

        return content

    sorted_test = sorted(concatenate_headings_content(('<h1>Test</h1>', '<h1>Test .*</h1>')))

    sorted_step = sorted(
        enumerate_content(concatenate_headings_content(('<h2>Step</h2>', '<h2>Test Step</h2>')))
        + sorted_test
    )
    step = ''.join([f"{v[1]}" for v in sorted_step])

    sorted_expect = sorted(
        enumerate_content(
            concatenate_headings_content(
                ('<h2>Expect</h2>', '<h2>Result</h2>', '<h2>Expected Result</h2>')
            )
        )
        + sorted_test
    )
    expect = ''.join([f"{v[1]}" for v in sorted_expect])

    def check_section_exists(text: str) -> str:
        try:
            return str(sections_headings[text][0][1])
        except (IndexError, KeyError):
            return ''

    setup = check_section_exists('<h1>Setup</h1>')
    cleanup = check_section_exists('<h1>Cleanup</h1>')

    return step, expect, setup, cleanup


def enabled_somewhere(test: 'tmt.Test') -> bool:
    """
    True if the test is enabled for some context (adjust rules)
    """

    # Already enabled, no need to dig deeper
    if test.enabled:
        return True
    # We need to find 'enabled' value before adjust happened
    # node.original_data are fmf data _before_ adjust was processed
    node = test.node
    enabled_not_set = True
    while enabled_not_set and node is not None:
        try:
            if node.original_data['enabled']:
                return True
            enabled_not_set = False
        except KeyError:
            pass
        # Not set in this node, check parent
        node = node.parent

    # Default value (True) of 'enabled' was used
    if enabled_not_set:
        return True

    # Some rule in adjust enables the test
    try:
        adjust_rules = test.node.original_data['adjust']
        # TODO: Should not be necessary once we normalize data
        if isinstance(adjust_rules, dict):
            adjust_rules = [adjust_rules]
        for rule in adjust_rules:
            try:
                if rule['enabled']:
                    return True
            except KeyError:
                pass
    except KeyError:
        pass
    # At this point nothing enables the test
    return False


def enabled_for_environment(test: 'tmt.base.Test', tcms_notes: str) -> bool:
    """
    Check whether test is enabled for specified environment
    """

    field = StructuredField(tcms_notes)
    context_dict = {}
    try:
        for line in cast(str, field.get('environment')).split('\n'):
            try:
                dimension, values = line.split('=', maxsplit=2)
                context_dict[dimension.strip()] = [
                    value.strip() for value in re.split(",|and", values)
                ]
            except ValueError:
                pass
    except tmt.utils.StructuredFieldError:
        pass

    if not context_dict:
        return True

    try:
        context = fmf.context.Context(**context_dict)
        test_node = test.node.copy()
        test_node.adjust(context, case_sensitive=False)
        return tmt.Test(node=test_node, logger=test._logger).enabled
    except BaseException as exception:
        log.debug(f"Failed to process adjust: {exception}")
        return True


def return_markdown_file() -> Optional[Path]:
    """
    Return path to the markdown file
    """

    files = '\n'.join(os.listdir())
    reg_exp = r'.+\.md$'
    md_files = re.findall(reg_exp, files, re.MULTILINE)
    fail_message = "in the current working directory.\nManual steps couldn't be exported"
    if len(md_files) == 1:
        return Path.cwd() / str(md_files[0])
    if not md_files:
        echo(style(f'Markdown file doesn\'t exist {fail_message}', fg='yellow'))
        return None

    echo(style(f'{len(md_files)} Markdown files found {fail_message}', fg='yellow'))
    return None


def get_category(path: Path) -> str:
    """
    Get category from Makefile
    """

    category = DEFAULT_NITRATE_CATEGORY
    try:
        category_search = re.search(
            r'echo\s+"Type:\s*(.*)"', (path / 'Makefile').read_text(encoding='utf-8'), re.MULTILINE
        )
        if category_search:
            category = category_search.group(1)
    # Default to 'Sanity' if Makefile or Type not found
    except (OSError, AttributeError):
        pass
    return category


def create_nitrate_case(summary: str, category: str) -> NitrateTestCase:
    """
    Create new nitrate case
    """

    # Create the new test case
    assert nitrate
    category = nitrate.Category(name=category, product=DEFAULT_PRODUCT)
    testcase: NitrateTestCase = nitrate.TestCase(summary=summary, category=category)
    echo(style(f"Test case '{testcase.identifier}' created.", fg='blue'))
    return testcase


def add_to_nitrate_runs(
    nitrate_case: NitrateTestCase,
    general_plan: NitrateTestPlan,
    test: 'tmt.Test',
    dry_mode: bool,
) -> None:
    """
    Add nitrate test case to all active runs under given general plan

    Go down plan tree from general plan, add case and case run to
    all open runs. Try to apply adjust.
    """

    assert nitrate
    for child_plan in nitrate.TestPlan.search(parent=general_plan.id):
        for testrun in child_plan.testruns:
            if testrun.status == nitrate.RunStatus("FINISHED"):
                continue
            if not enabled_for_environment(test, tcms_notes=testrun.notes):
                continue
            # nitrate_case is None when --dry and --create are used together
            if not nitrate_case or child_plan not in nitrate_case.testplans:
                echo(style(f"Link to plan '{child_plan}'.", fg='magenta'))
                if not dry_mode:
                    nitrate_case.testplans.add(child_plan)
            if not nitrate_case or nitrate_case not in [caserun.testcase for caserun in testrun]:
                echo(style(f"Link to run '{testrun}'.", fg='magenta'))
                if not dry_mode:
                    nitrate.CaseRun(testcase=nitrate_case, testrun=testrun)


def prepare_extra_summary(test: 'tmt.Test', append_summary: bool) -> str:
    """
    extra-summary for export --create test
    """

    assert test.fmf_id.url is not None  # narrow type

    parsed_url = urllib.parse.urlparse(test.fmf_id.url)
    remote_dirname = re.sub('.git$', '', Path(parsed_url.path).name)
    if not remote_dirname:
        raise ConvertError("Unable to find git remote url.")
    generated = f"{remote_dirname} {test.name}"
    if test.summary and append_summary:
        generated += f" - {test.summary}"
    # FIXME: cast() - no issue, type-less "dispatcher" method
    return cast(str, test.node.get('extra-summary', generated))


# avoid multiple searching for general plans (it is expensive)
@cache
def find_general_plan(component: str) -> NitrateTestPlan:
    """
    Return single General Test Plan or raise an error
    """

    assert nitrate
    # At first find by linked components
    found: list[NitrateTestPlan] = nitrate.TestPlan.search(
        type__name="General",
        is_active=True,
        component__name=f"{component}",
    )
    # Attempt to find by name if no test plan found
    if not found:
        found = nitrate.TestPlan.search(
            type__name="General",
            is_active=True,
            name=f"{component} / General",
        )
    # No general -> raise error
    if not found:
        raise nitrate.NitrateError(f"No general test plan found for '{component}'.")
    # Multiple general plans are fishy -> raise error
    if len(found) != 1:
        nitrate.NitrateError("Multiple general test plans found for '{component}' component.")
    # Finally return the one and only General plan
    return found[0]


def export_to_nitrate(test: 'tmt.Test') -> None:
    """
    Export fmf metadata to nitrate test cases
    """

    import tmt.base

    import_nitrate()
    assert nitrate
    assert gssapi

    # Check command line options
    create = test.opt('create')
    general = test.opt('general')
    link_runs = test.opt('link_runs')
    duplicate = test.opt('duplicate')
    link_bugzilla = test.opt('bugzilla')
    ignore_git_validation = test.opt('ignore_git_validation')
    dry_mode = test.is_dry_run
    append_summary = test.opt('append-summary')

    if link_runs:
        general = True

    # Check git is already correct
    valid, error_msg = tmt.utils.git.validate_git_status(test)
    if not valid:
        if ignore_git_validation:
            echo(style(f"Exporting regardless '{error_msg}'.", fg='red'))
        else:
            raise ConvertError(
                f"Can't export due '{error_msg}'.\n"
                "Use --ignore-git-validation on your own risk to export regardless."
            )

    # Check nitrate test case
    try:
        nitrate_id = test.node.get('extra-nitrate')[3:]
        nitrate_case: NitrateTestCase = nitrate.TestCase(int(nitrate_id))
        nitrate_case.summary  # noqa: B018 - Make sure we connect to the server now
        echo(style(f"Test case '{nitrate_case.identifier}' found.", fg='blue'))
    except TypeError:
        # Create a new nitrate test case
        if create:
            nitrate_case = None
            # Check for existing Nitrate tests with the same fmf id
            if not duplicate:
                testcases = _nitrate_find_fmf_testcases(test)
                with suppress(StopIteration):
                    nitrate_case = next(testcases)
            if not nitrate_case:
                # Summary for TCMS case
                extra_summary = prepare_extra_summary(test, append_summary)
                assert test.path is not None  # narrow type
                category = get_category(test.node.root / test.path.unrooted())
                if not dry_mode:
                    nitrate_case = create_nitrate_case(extra_summary, category)
                else:
                    echo(style(f"Test case '{extra_summary}' created.", fg='blue'))
                test._metadata['extra-summary'] = extra_summary
            # Either newly created or duplicate with missing extra-nitrate
            if nitrate_case:
                echo(style("Append the nitrate test case id.", fg='green'))
                if not dry_mode:
                    with test.node as data:
                        data["extra-nitrate"] = nitrate_case.identifier
        else:
            raise ConvertError(
                f"Nitrate test case id not found for {test}"
                " (You can use --create option to enforce"
                " creating testcases)"
            )
    except (nitrate.NitrateError, gssapi.raw.misc.GSSError) as error:
        raise ConvertError(error)

    # Check if URL is accessible, to be able to reach from nitrate
    tmt.utils.git.check_git_url(test.fmf_id.url, test._logger)

    # Summary
    try:
        summary = (
            test._metadata.get('extra-summary')
            or test._metadata.get('extra-task')
            or prepare_extra_summary(test, append_summary)
        )
    except ConvertError:
        summary = test.name
    if not dry_mode:
        nitrate_case.summary = summary
    echo(style('summary: ', fg='green') + summary)

    # Script
    if test.node.get('extra-task'):
        if not dry_mode:
            nitrate_case.script = test.node.get('extra-task')
        echo(style('script: ', fg='green') + test.node.get('extra-task'))

    # Components and General plan
    # First remove any components that are already there
    if not dry_mode:
        nitrate_case.components.clear()
    # Only these general plans should stay
    expected_general_plans = set()
    # Then add fmf ones
    if test.component:
        echo(style('components: ', fg='green') + ' '.join(test.component))
        for component in test.component:
            try:
                nitrate_component = nitrate.Component(name=component, product=DEFAULT_PRODUCT.id)
                if not dry_mode:
                    nitrate_case.components.add(nitrate_component)
            except nitrate.xmlrpc_driver.NitrateError as error:
                log.debug(error)
                echo(style(f"Failed to add component '{component}'.", fg='red'))
            if general:
                try:
                    general_plan = find_general_plan(component)
                    expected_general_plans.add(general_plan)
                    echo(style(f"Linked to general plan '{general_plan}'.", fg='magenta'))
                    if not dry_mode:
                        nitrate_case.testplans.add(general_plan)
                    if link_runs:
                        add_to_nitrate_runs(nitrate_case, general_plan, test, dry_mode)
                except nitrate.NitrateError as error:
                    log.debug(error)
                    echo(style(f"Failed to find general test plan for '{component}'.", fg='red'))
    # Remove unexpected general plans
    if general and nitrate_case:
        # Remove also all general plans linked to testcase
        for nitrate_plan in list(nitrate_case.testplans):
            if nitrate_plan.type.name == "General" and nitrate_plan not in expected_general_plans:
                echo(style(f"Removed general plan '{nitrate_plan}'.", fg='red'))
                if not dry_mode:
                    nitrate_case.testplans.remove(nitrate_plan)

    # Tags
    # Convert 'tier' attribute into a Tier tag
    if test.tier is not None:
        test.tag.append(f"Tier{test.tier}")
    # Add special fmf-export tag
    test.tag.append('fmf-export')
    if not dry_mode:
        nitrate_case.tags.clear()
        nitrate_case.tags.add([nitrate.Tag(tag) for tag in test.tag])
    echo(style('tags: ', fg='green') + ' '.join(set(test.tag)))

    # Default tester
    if test.contact:
        try:
            # Need to pick one value, so picking the first contact
            email_address = email.utils.parseaddr(test.contact[0])[1]
            nitrate_user = nitrate.User(email_address)
            nitrate_user._fetch()  # To check that user exists
            if not dry_mode:
                nitrate_case.tester = nitrate_user
            echo(style('default tester: ', fg='green') + email_address)
        except nitrate.NitrateError as error:
            log.debug(error)
            raise ConvertError(f"Nitrate issue: {error}")

    # Duration
    if not dry_mode:
        nitrate_case.time = test.duration
    echo(style('estimated time: ', fg='green') + test.duration)

    # Manual
    if not dry_mode:
        nitrate_case.automated = not test.manual
    echo(style('automated: ', fg='green') + ['auto', 'manual'][test.manual])

    # Status
    current_status = nitrate_case.status if nitrate_case else nitrate.CaseStatus('CONFIRMED')
    # Enable enabled tests
    if enabled_somewhere(test):
        if not dry_mode:
            nitrate_case.status = nitrate.CaseStatus('CONFIRMED')
        echo(style('status: ', fg='green') + 'CONFIRMED')
    # Disable disabled tests which are CONFIRMED
    elif current_status == nitrate.CaseStatus('CONFIRMED'):
        if not dry_mode:
            nitrate_case.status = nitrate.CaseStatus('DISABLED')
        echo(style('status: ', fg='green') + 'DISABLED')
    # Keep disabled tests in their states
    else:
        echo(style('status: ', fg='green') + str(current_status))

    # Environment
    if test.environment:
        environment = ' '.join(tmt.utils.shell_variables(test.environment))
        if not dry_mode:
            nitrate_case.arguments = environment
        echo(style('arguments: ', fg='green') + environment)
    else:
        # FIXME unable clear to set empty arguments
        # (possibly error in xmlrpc, BZ#1805687)
        if not dry_mode:
            nitrate_case.arguments = ' '
        echo(style('arguments: ', fg='green') + "' '")

    # Structured Field
    struct_field = StructuredField(nitrate_case.notes if nitrate_case else '')
    echo(style('Structured Field: ', fg='green'))

    # Mapping of structured field sections to fmf case attributes
    section_to_attr = {
        'description': test.summary,
        'purpose-file': test.description,
        'hardware': test.node.get('extra-hardware'),
        'pepa': test.node.get('extra-pepa'),
    }
    for section, attribute in section_to_attr.items():
        if attribute is None:
            with suppress(tmt.utils.StructuredFieldError):
                struct_field.remove(section)
        else:
            struct_field.set(section, attribute)
            echo(style(section + ': ', fg='green') + attribute.strip())

    # fmf identifier
    fmf_id = tmt.utils.dict_to_yaml(test.fmf_id.to_minimal_spec())
    struct_field.set('fmf', fmf_id)
    echo(style('fmf id:\n', fg='green') + fmf_id.strip())

    # Warning
    if WARNING not in struct_field.header():
        struct_field.header(WARNING + struct_field.header())
        echo(style('Add migration warning to the test case notes.', fg='green'))

    # ID
    uuid = tmt.identifier.add_uuid_if_not_defined(test.node, dry_mode, test._logger)
    if not uuid:
        uuid = test.node.get(tmt.identifier.ID_KEY)
    struct_field.set(tmt.identifier.ID_KEY, uuid)
    echo(style(f"Append the ID {uuid}.", fg='green'))

    # Saving case.notes with edited StructField
    if not dry_mode:
        nitrate_case.notes = struct_field.save()

    # Export manual test instructions from *.md file to nitrate as html
    md_path = return_markdown_file()
    if md_path and md_path.exists():
        step, expect, setup, cleanup = convert_manual_to_nitrate(md_path)
        if not dry_mode:
            nitrate.User()._server.TestCase.store_text(
                nitrate_case.id, step, expect, setup, cleanup
            )
        echo(style("manual steps:", fg='green') + f" found in {md_path}")

    # List of bugs test verifies
    verifies_bug_ids = []
    if test.link:
        for link in test.link.get('verifies'):
            if isinstance(link.target, tmt.base.FmfId):
                log.debug(f"Will not look for bugzila URL in fmf id '{link.target}'.")
                continue

            try:
                bug_id_search = re.search(tmt.export.RE_BUGZILLA_URL, link.target)
                if not bug_id_search:
                    log.debug(f"Did not find bugzila URL in '{link.target}'.")
                    continue
                bug_id = int(bug_id_search.group(1))
                verifies_bug_ids.append(bug_id)
            except Exception as err:
                log.debug(err)

    # Add bugs to the Nitrate case
    if verifies_bug_ids:
        echo(
            style('Verifies bugs: ', fg='green') + ', '.join([f"BZ#{b}" for b in verifies_bug_ids])
        )
    for bug_id in verifies_bug_ids:
        if not dry_mode:
            nitrate_case.bugs.add(nitrate.Bug(bug=int(bug_id)))

    # Update nitrate test case
    if not dry_mode:
        nitrate_case.update()
        echo(
            style(
                f"Test case '{nitrate_case.identifier}' successfully exported to nitrate.",
                fg='magenta',
            )
        )

    # Optionally link Bugzilla to Nitrate case
    if link_bugzilla and verifies_bug_ids and not dry_mode:
        tmt.export.bz_set_coverage(verifies_bug_ids, nitrate_case.id, NITRATE_TRACKER_ID)


@tmt.base.Test.provides_export('nitrate')
class NitrateExporter(tmt.export.ExportPlugin):
    @classmethod
    def export_test_collection(
        cls, tests: list[tmt.base.Test], keys: Optional[list[str]] = None, **kwargs: Any
    ) -> str:
        for test in tests:
            export_to_nitrate(test)

        return ''



#!/usr/bin/env python3

import os
import sys

step = os.getenv('STEP', '')

isatty = (sys.stdin.isatty(), sys.stdout.isatty(), sys.stderr.isatty())

print(f'{step}: stdin:', isatty[0])
print(f'{step}: stdout:', isatty[1])
print(f'{step}: stderr:', isatty[2])

sys.exit(0 if not any(isatty) else 1)



import functools
from typing import Any, Optional

import tmt
import tmt.config
import tmt.utils
import tmt.utils.templates
from tmt.utils import Path

DEFAULT_CUSTOM_TEMPLATES_PATH = tmt.config.effective_config_dir() / 'templates'
DEFAULT_PLAN_NAME = "/default/plan"
INIT_TEMPLATES = ['mini', 'base', 'full']
TEMPLATE_FILE_SUFFIX = '.j2'
TEMPLATE_TYPES = ['default', 'story', 'plan', 'test', 'script']

# TemplatesType is a dictionary of template types and their paths.
# It follows the following structure: {template_type: {template_name: template_path}}
# e.g. {'story': {'mini': Path('templates/story/mini.j2')}, ...}
TemplatesType = dict[str, dict[str, Path]]


def _combine(default: TemplatesType, custom: TemplatesType) -> TemplatesType:
    """
    Combines default templates and custom templates.
    Custom templates have priority and potentially override default templates.
    """

    result: TemplatesType = {}
    for key in default:
        result[key] = {**default[key], **custom.get(key, {})}
    return result


def _get_template_file_paths(path: Path) -> dict[str, Path]:
    """
    Get a dictionary of template names and their file paths.
    :param path: Path to the directory to search for templates.
    """

    return {
        file.name.removesuffix(TEMPLATE_FILE_SUFFIX): file
        for file in path.iterdir()
        if file.is_file() and file.suffix == TEMPLATE_FILE_SUFFIX
    }


def _get_templates(root_dir: Path) -> TemplatesType:
    """
    Get all templates in given root directory.
    :param root_dir: Path to the directory to search for templates.
    """

    templates: TemplatesType = {}
    for template_type in TEMPLATE_TYPES:
        templates_dir = root_dir / template_type
        if templates_dir.exists() and templates_dir.is_dir():
            template_files = _get_template_file_paths(templates_dir)
            if template_files:
                templates[template_type] = template_files
    return templates


def _append_newline_if_missing(input_string: str) -> str:
    """
    Append newline to the input if it doesn't end with one.
    """

    return input_string if input_string.endswith('\n') else input_string + '\n'


class TemplateManager:
    """
    Template manager class.

    It provides methods for rendering templates during story, plan or test creation.
    """

    def __init__(self, custom_template_path: Optional[Path] = None):
        self.custom_template_path = custom_template_path or DEFAULT_CUSTOM_TEMPLATES_PATH
        self._init_custom_templates_folder()
        self._environment = tmt.utils.templates.default_template_environment()

    @functools.cached_property
    def templates(self) -> TemplatesType:
        """
        Return all available templates (default and optional).
        """

        return _combine(self.default_templates, self.custom_templates)

    @functools.cached_property
    def default_templates(self) -> TemplatesType:
        """
        Return all default templates.
        """

        templates_dir = tmt.utils.resource_files('templates/')
        templates = _get_templates(templates_dir)
        if not templates:
            raise tmt.utils.GeneralError(f"Could not find default templates in '{templates_dir}'.")
        return templates

    @functools.cached_property
    def custom_templates(self) -> TemplatesType:
        """
        Return all custom templates.
        """

        return _get_templates(self.custom_template_path)

    def render_default_plan(self) -> str:
        """
        Return default plan template.
        """

        try:
            path = self.default_templates['default']['plan']
        except KeyError:
            raise tmt.utils.GeneralError("Default plan template not found.")

        return _append_newline_if_missing(self.render_file(path, plan_name=DEFAULT_PLAN_NAME))

    def render_from_url(self, url: str, **variables: Any) -> str:
        """
        Render template from given URL.
        :param url: URL to the template file.
        :param variables: variables to be passed to the template.
        """

        template = tmt.utils.get_url_content(url)
        template = tmt.utils.templates.render_template(
            template, None, self._environment, **variables
        )
        return _append_newline_if_missing(template)

    def render_file(self, path: Path, **variables: Any) -> str:
        """
        Render template from given file path.
        :param path: path to the template file.
        :param variables: variables to be passed to the template.
        """

        template = tmt.utils.templates.render_template_file(path, self._environment, **variables)
        return _append_newline_if_missing(template)

    def _init_custom_templates_folder(self) -> None:
        """
        Create custom template folders if they don't exist.
        """

        for key in TEMPLATE_TYPES:
            path = self.custom_template_path / key
            try:
                path.mkdir(parents=True, exist_ok=True)
            except OSError as error:
                raise tmt.utils.GeneralError(
                    f"Failed to create template folder '{path}'.\n{error}"
                ) from error


# Global TemplateManager object
MANAGER = TemplateManager()



import xml.dom
import xml.dom.minidom
from unittest.mock import MagicMock, PropertyMock

import pytest

from tmt.result import Result, ResultOutcome
from tmt.steps.report.junit import ReportJUnit, ReportJUnitData
from tmt.utils import Path


@pytest.fixture
def report_fix(tmppath: Path, root_logger):
    # need to provide genuine workdir paths - mock would break os.path.* calls
    step_mock = MagicMock(workdir=tmppath)
    plan_mock = MagicMock()
    name_property = PropertyMock(return_value='name')

    type(plan_mock).name = name_property
    type(step_mock).plan = plan_mock

    out_file_path = Path(tmppath / "out.xml")

    report = ReportJUnit(
        logger=root_logger,
        step=step_mock,
        data=ReportJUnitData(name='x', how='junit', file=out_file_path),
        workdir=tmppath / 'junit',
    )
    report.info = MagicMock()

    results = []

    execute_mock = MagicMock()
    type(execute_mock).results = MagicMock(return_value=results)
    type(plan_mock).execute = execute_mock

    return report, results, out_file_path


# A simple way would be comparing XML nodes as strings, e.g. output of their
# .toprettyxml() methods. But that has been observed as not fully deterministic
# with Python 3.6, changing order or attributes from time to time. Therefore
# taking the longer, more verbose approach.
def _compare_xml_node(tree_path: list[str], expected: xml.dom.Node, actual: xml.dom.Node) -> None:
    """
    Assert two XML nodes have the same content
    """

    # All of this would be doable in a much, much simpler, condensed manner,
    # but being more verbose allows for way more specific error message when
    # comparison fails.

    tree_path_joined = '.'.join(tree_path)

    # Make sure node names do match.
    assert expected.nodeName == actual.nodeName, (
        f"Element name mismatch at {tree_path_joined}: "
        f"expected {expected.nodeName}, found {actual.nodeName}"
    )

    # If nodes have the same tag, move on to attributes. Make sure both nodes
    # have the same set of attributes, with same respective values.
    #
    # Note: sometimes, node.attributes may be set to `None`, meaning "no attributes",
    # which is fine as long as the other node also has no attributes.
    expected_attributes = sorted((expected.attributes or {}).items())
    actual_attributes = sorted((actual.attributes or {}).items())

    assert len(expected_attributes) == len(actual_attributes), (
        f"Attribute count mismatch at {tree_path_joined}: "
        f"expected {len(expected_attributes)}, found {len(actual_attributes)}"
    )

    for (expected_name, expected_value), (actual_name, actual_value) in zip(
        expected_attributes, actual_attributes
    ):
        assert expected_name == actual_name, (
            f'Attribute mismatch at {tree_path_joined}: '
            f'expected {expected_name}="{expected_value}"'
        )
        assert expected_value == actual_value, (
            f'Attribute mismatch at {tree_path_joined}: found {actual_name}="{actual_value}"'
        )

    # Hooray, attributes match. Dig deeper, how about children?
    # To compare children, use this very function to compare each child with
    # a corresponding child of the other node.
    #
    # Note: skip empty text nodes. For the purpose of comparing XML structure of junit
    # XML, they are not important - in this case, they are spawned by indentation of
    # elements in the expected XML string, and they do not affect the semantics of those
    # nodes.
    def _valid_children(node: xml.dom.Node) -> list[xml.dom.Node]:
        return [
            child
            for child in node.childNodes
            if child.nodeType != xml.dom.Node.TEXT_NODE or child.data.strip()
        ]

    expected_children = _valid_children(expected)
    actual_children = _valid_children(actual)

    assert len(expected_children) == len(actual_children), (
        f"Children count mismatch at {tree_path_joined}: "
        f"expected {len(expected_children)}, found {len(actual_children)}"
    )

    return all(
        _compare_xml_node([*tree_path, expected_child.nodeName], expected_child, actual_child)
        for expected_child, actual_child in zip(expected.childNodes, actual.childNodes)
    )


def assert_xml(actual_filepath, expected):
    with (
        open(actual_filepath) as f,
        xml.dom.minidom.parse(f) as actual_dom,
        xml.dom.minidom.parseString(expected) as expected_dom,
    ):
        assert _compare_xml_node([expected_dom.nodeName], expected_dom, actual_dom)


@pytest.mark.skipif(pytest.__version__.startswith('3'), reason="too old pytest")
class TestStateMapping:
    def test_pass(self, report_fix):
        report, results, out_file_path = report_fix
        results.append(Result(result=ResultOutcome.PASS, name="/pass", serial_number=1))

        report.go()

        assert_xml(
            out_file_path,
            """<?xml version='1.0' encoding='utf-8'?>
<testsuites disabled="0" errors="0" failures="0" tests="1" time="0.0">
  <testsuite name="name" disabled="0" errors="0" failures="0" skipped="0" tests="1" time="0.0">
    <testcase name="/pass">


                </testcase>
  </testsuite>
</testsuites>
""",
        )

    def test_info(self, report_fix):
        report, results, out_file_path = report_fix
        results.append(Result(result=ResultOutcome.INFO, name="/info", serial_number=1))
        report.go()

        assert_xml(
            out_file_path,
            """<?xml version='1.0' encoding='utf-8'?>
<testsuites disabled="0" errors="0" failures="0" tests="1" time="0.0">
  <testsuite name="name" disabled="0" errors="0" failures="0" skipped="1" tests="1" time="0.0">
    <testcase name="/info">
      <skipped type="skipped" message="info"/>
    </testcase>
  </testsuite>
</testsuites>
""",
        )

    def test_warn(self, report_fix):
        report, results, out_file_path = report_fix
        results.append(Result(result=ResultOutcome.WARN, name="/warn", serial_number=1))
        report.go()

        assert_xml(
            out_file_path,
            """<?xml version='1.0' encoding='utf-8'?>
<testsuites disabled="0" errors="1" failures="0" tests="1" time="0.0">
  <testsuite name="name" disabled="0" errors="1" failures="0" skipped="0" tests="1" time="0.0">
    <testcase name="/warn">
      <error type="error" message="warn"/>
    </testcase>
  </testsuite>
</testsuites>
""",
        )

    def test_error(self, report_fix):
        report, results, out_file_path = report_fix
        results.append(Result(result=ResultOutcome.ERROR, name="/error", serial_number=1))
        report.go()

        assert_xml(
            out_file_path,
            """<?xml version='1.0' encoding='utf-8'?>
<testsuites disabled="0" errors="1" failures="0" tests="1" time="0.0">
  <testsuite name="name" disabled="0" errors="1" failures="0" skipped="0" tests="1" time="0.0">
    <testcase name="/error">
      <error type="error" message="error"/>
    </testcase>
  </testsuite>
</testsuites>
""",
        )

    def test_fail(self, report_fix):
        report, results, out_file_path = report_fix
        results.append(Result(result=ResultOutcome.FAIL, name="/fail", serial_number=1))
        report.go()

        assert_xml(
            out_file_path,
            """<?xml version='1.0' encoding='utf-8'?>
<testsuites disabled="0" errors="0" failures="1" tests="1" time="0.0">
  <testsuite name="name" disabled="0" errors="0" failures="1" skipped="0" tests="1" time="0.0">
    <testcase name="/fail">
      <failure type="failure" message="fail"/>
    </testcase>
  </testsuite>
</testsuites>
""",
        )



import logging
from typing import cast
from unittest.mock import MagicMock

import _pytest.logging
import pytest
from testcloud.domain_configuration import DomainConfiguration, TPMConfiguration

from tests.unit import MATCH, assert_log
from tmt.hardware import TPM_VERSION_ALLOWED_OPERATORS, Hardware, Operator
from tmt.log import Logger
from tmt.steps.provision.testcloud import (
    TPM_VERSION_ALLOWED_OPERATORS as virtual_TPM_VERSION_ALLOWED_OPERATORS,  # noqa: N811
)
from tmt.steps.provision.testcloud import (
    _apply_hw_tpm,
    import_testcloud,
)

import_testcloud(Logger.get_bootstrap_logger())

# E402: these must be imported *after* importing testcloud
from tmt.steps.provision.testcloud import (  # noqa: E402
    TPM_CONFIG_ALLOWS_VERSIONS,
    TPM_VERSION_SUPPORTED_VERSIONS,
)

if TPM_CONFIG_ALLOWS_VERSIONS:
    allowed_combinations: list[tuple[str, Operator]] = []

    for version in TPM_VERSION_SUPPORTED_VERSIONS[TPM_CONFIG_ALLOWS_VERSIONS]:
        for op in virtual_TPM_VERSION_ALLOWED_OPERATORS:
            allowed_combinations.append((version, op))

    @pytest.mark.parametrize(
        ('version', 'op'),
        allowed_combinations,
        ids=[f'{op.value} {version}' for version, op in allowed_combinations],
    )
    def test_tpm(
        root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture, version: str, op: Operator
    ) -> None:
        mock_domain = MagicMock(name='<domain>')

        _apply_hw_tpm(
            Hardware.from_spec({'tpm': {'version': f'{op.value} {version}'}}),
            mock_domain,
            root_logger,
        )

        tpm_config = cast(DomainConfiguration, mock_domain).tpm_configuration

        assert isinstance(tpm_config, TPMConfiguration)

        assert tpm_config.version == version

        assert_log(
            caplog,
            message=MATCH(
                rf"tpm.version: set to '{version}' because of 'tpm.version: {op.value} {version}'"
            ),
            levelno=logging.DEBUG,
        )

else:
    allowed_combinations: list[tuple[str, Operator]] = []

    for version in TPM_VERSION_SUPPORTED_VERSIONS[TPM_CONFIG_ALLOWS_VERSIONS]:
        for op in virtual_TPM_VERSION_ALLOWED_OPERATORS:
            allowed_combinations.append((version, op))

    @pytest.mark.parametrize(
        ('version', 'op'),
        allowed_combinations,
        ids=[f'{op.value} {version}' for version, op in allowed_combinations],
    )
    def test_tpm_with_default_version(
        root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture, version: str, op: Operator
    ) -> None:
        mock_domain = MagicMock(name='<domain>')

        _apply_hw_tpm(
            Hardware.from_spec({'tpm': {'version': f'{op.value} {version}'}}),
            mock_domain,
            root_logger,
        )

        tpm_config = cast(DomainConfiguration, mock_domain).tpm_configuration

        assert isinstance(tpm_config, TPMConfiguration)

        assert_log(
            caplog,
            message=MATCH(
                rf"tpm.version: set to '{version}' because of 'tpm.version: {op.value} {version}'"
            ),
            levelno=logging.DEBUG,
        )


def test_tpm_no_hardware(root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture) -> None:
    mock_domain = MagicMock(name='<domain>')

    _apply_hw_tpm(None, mock_domain, root_logger)
    assert cast(DomainConfiguration, mock_domain).tpm_configuration is None

    assert_log(
        caplog,
        message=MATCH(r"tpm.version: not included because of no constraints"),
        levelno=logging.DEBUG,
    )


def test_tpm_no_hardware_constraint(
    root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture
) -> None:
    mock_domain = MagicMock(name='<domain>')

    _apply_hw_tpm(Hardware(constraint=None, spec=None), mock_domain, root_logger)
    assert cast(DomainConfiguration, mock_domain).tpm_configuration is None

    assert_log(
        caplog,
        message=MATCH(r"tpm.version: not included because of no constraints"),
        levelno=logging.DEBUG,
    )


def test_tpm_no_tpm_constraints(
    root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture
) -> None:
    mock_domain = MagicMock(name='<domain>')

    _apply_hw_tpm(Hardware.from_spec({'memory': '4 GB'}), mock_domain, root_logger)

    assert cast(DomainConfiguration, mock_domain).tpm_configuration is None

    assert_log(
        caplog,
        message=MATCH(r"tpm.version: not included because of no 'tpm.version' constraints"),
        levelno=logging.DEBUG,
    )


def test_tpm_unsupported_version(
    root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture
) -> None:
    mock_domain = MagicMock(name='<domain>')

    _apply_hw_tpm(Hardware.from_spec({'tpm': {'version': '0.0.0'}}), mock_domain, root_logger)

    assert cast(DomainConfiguration, mock_domain).tpm_configuration is None

    assert_log(
        caplog,
        message=MATCH(
            r"warn: Cannot apply hardware requirement 'tpm\.version: == 0\.0\.0', "
            r"TPM version not supported."
        ),
        levelno=logging.WARNING,
    )


@pytest.mark.parametrize(
    'op',
    [
        op.value
        for op in TPM_VERSION_ALLOWED_OPERATORS
        if op not in virtual_TPM_VERSION_ALLOWED_OPERATORS
    ],
)
def test_tpm_unsupported_operator(
    root_logger: Logger, caplog: _pytest.logging.LogCaptureFixture, op: str
) -> None:
    mock_domain = MagicMock(name='<domain>')

    _apply_hw_tpm(Hardware.from_spec({'tpm': {'version': f'{op} 2.0'}}), mock_domain, root_logger)

    assert cast(DomainConfiguration, mock_domain).tpm_configuration is None

    assert_log(
        caplog,
        message=MATCH(
            rf"warn: Cannot apply hardware requirement 'tpm\.version: {op} 2\.0', operator not supported."  # noqa: E501
        ),
        levelno=logging.WARNING,
    )



import datetime
import os
from typing import Optional

from requests import post

import tmt
import tmt.steps
import tmt.steps.report
import tmt.utils
from tmt.container import container, field
from tmt.utils import Path

from .junit import ResultsContext, make_junit_xml

DEFAULT_FILENAME = 'xunit.xml'


@container
class ReportPolarionData(tmt.steps.report.ReportStepData):
    file: Optional[Path] = field(
        default=None,
        option='--file',
        metavar='FILE',
        help='Path to the file to store xUnit in.',
        normalize=tmt.utils.normalize_path,
    )

    upload: bool = field(
        default=True,
        option=('--upload / --no-upload'),
        is_flag=True,
        show_default=True,
        help="""
            Whether to upload results to Polarion,
            also uses environment variable TMT_PLUGIN_REPORT_POLARION_UPLOAD.
            """,
    )

    project_id: Optional[str] = field(
        default=None,
        option='--project-id',
        metavar='ID',
        help="""
             Use specific Polarion project ID,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_PROJECT_ID. If no project ID
             is found, the project ID is taken from pylero configuration default project setting as
             a last resort.
             """,
    )

    title: Optional[str] = field(
        default=None,
        option='--title',
        metavar='TITLE',
        help="""
             Use specific test run title,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_TITLE.
             """,
    )

    description: Optional[str] = field(
        default=None,
        option='--description',
        metavar='DESCRIPTION',
        help="""
             Use specific test run description,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_DESCRIPTION.
             """,
    )

    template: Optional[str] = field(
        default=None,
        option='--template',
        metavar='TEMPLATE',
        help="""
             Use specific test run template,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_TEMPLATE.
             """,
    )

    use_facts: bool = field(
        default=False,
        option=('--use-facts / --no-use-facts'),
        is_flag=True,
        show_default=True,
        help="""
            Use hostname and arch from guest facts,
            also uses environment variable TMT_PLUGIN_REPORT_POLARION_USE_FACTS.
            """,
    )

    planned_in: Optional[str] = field(
        default=None,
        option='--planned-in',
        metavar='PLANNEDIN',
        help="""
             Select a specific release to mark this test run with,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_PLANNED_IN.
             """,
    )

    assignee: Optional[str] = field(
        default=None,
        option='--assignee',
        metavar='ASSIGNEE',
        help="""
             Who is responsible for this test run,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_ASSIGNEE.
             """,
    )

    pool_team: Optional[str] = field(
        default=None,
        option='--pool-team',
        metavar='POOLTEAM',
        help="""
             Which subsystem is this test run relevant for,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_POOL_TEAM.
             """,
    )

    arch: Optional[str] = field(
        default=None,
        option='--arch',
        metavar='ARCH',
        help="""
             Which architecture was this run executed on,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_ARCH.
             """,
    )

    platform: Optional[str] = field(
        default=None,
        option='--platform',
        metavar='PLATFORM',
        help="""
             Which platform was this run executed on,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_PLATFORM.
             """,
    )

    build: Optional[str] = field(
        default=None,
        option='--build',
        metavar='BUILD',
        help="""
             Which build was this run executed on,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_BUILD.
             """,
    )

    sample_image: Optional[str] = field(
        default=None,
        option='--sample-image',
        metavar='SAMPLEIMAGE',
        help="""
             Which sample image was this run executed on,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_SAMPLE_IMAGE.
             """,
    )

    logs: Optional[str] = field(
        default=None,
        option='--logs',
        metavar='LOGLOCATION',
        help="""
             Location of the logs for this test run,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_LOGS.
             Ultimately also uses environment variable TMT_REPORT_ARTIFACTS_URL.
             """,
    )

    compose_id: Optional[str] = field(
        default=None,
        option='--compose-id',
        metavar='COMPOSEID',
        help="""
             Compose ID of image used for this run,
             also uses environment variable TMT_PLUGIN_REPORT_POLARION_COMPOSE_ID.
             """,
    )

    fips: bool = field(
        default=False,
        option=('--fips / --no-fips'),
        is_flag=True,
        show_default=True,
        help='FIPS mode enabled or disabled for this run.',
    )

    prettify: bool = field(
        default=True,
        option=('--prettify / --no-prettify'),
        is_flag=True,
        show_default=True,
        help="Enable the XML pretty print for generated XUnit file.",
    )

    include_output_log: bool = field(
        default=True,
        option=('--include-output-log / --no-include-output-log'),
        is_flag=True,
        show_default=True,
        help='Include full standard output in resulting xml file.',
    )


@tmt.steps.provides_method('polarion')
class ReportPolarion(tmt.steps.report.ReportPlugin[ReportPolarionData]):
    """
    Write test results into an xUnit file and upload to Polarion.

    In order to get quickly started create a pylero config
    file ``~/.pylero`` in your home directory with the
    following content:

    .. code-block:: ini

        [webservice]
        url=https://{your polarion web URL}/polarion
        svn_repo=https://{your polarion web URL}/repo
        default_project={your project name}
        user={your username}
        password={your password}

    See the ``Pylero Documentation`` for more details on how
    to configure the ``pylero`` module.

    https://github.com/RedHatQE/pylero

    .. note::

        For Polarion report to export correctly you need to
        use password authentication, since exporting the
        report happens through Polarion XUnit importer which
        does not support using tokens. You can still
        authenticate with token to only generate the report
        using ``--no-upload`` argument.

    .. note::

        Your Polarion project might need a custom value format
        for the ``arch``, ``planned-in`` and other fields. The
        format of these fields might differ across Polarion
        projects, for example, ``x8664`` can be used instead
        of ``x86_64`` for the architecture.

    Examples:

    .. code-block:: yaml

        # Enable polarion report from the command line
        tmt run --all report --how polarion --project-id tmt
        tmt run --all report --how polarion --project-id tmt --no-upload --file test.xml

    .. code-block:: yaml

        # Use polarion as the default report for given plan
        report:
            how: polarion
            file: test.xml
            project-id: tmt
            title: tests_that_pass
            planned-in: RHEL-9.1.0
            pool-team: sst_tmt
    """

    _data_class = ReportPolarionData

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        members = super()._preserved_workdir_members

        if self.data.file is None:
            members = {*members, DEFAULT_FILENAME}

        return members

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Go through executed tests and report into Polarion
        """

        super().go(logger=logger)

        from tmt.export.polarion import find_polarion_case_ids, import_polarion

        import_polarion()
        from tmt.export.polarion import PolarionWorkItem

        title = self.data.title
        if not title:
            title = os.getenv(
                'TMT_PLUGIN_REPORT_POLARION_TITLE',
                self.step.plan.name.rsplit('/', 1)[1]
                + '_'
                +
                # Polarion server running with UTC timezone
                datetime.datetime.now(tz=datetime.timezone.utc).strftime("%Y%m%d%H%M%S"),
            )

        title = title.replace('-', '_')
        template = self.data.template or os.getenv('TMT_PLUGIN_REPORT_POLARION_TEMPLATE')
        project_id = self.data.project_id or os.getenv(
            'TMT_PLUGIN_REPORT_POLARION_PROJECT_ID', PolarionWorkItem._session.default_project
        )

        # The project_id is required
        if not project_id:
            raise tmt.utils.ReportError(
                "The Polarion project ID could not be determined. Consider setting it using "
                "'--project-id' argument or by setting 'TMT_PLUGIN_REPORT_POLARION_PROJECT_ID' "
                "environment variable."
            )

        # TODO: try use self.data instead - but these fields are not optional, they do have
        # default values, do envvars even have any effect at all??
        upload = self.get('upload', os.getenv('TMT_PLUGIN_REPORT_POLARION_UPLOAD'))
        use_facts = self.get('use-facts', os.getenv('TMT_PLUGIN_REPORT_POLARION_USE_FACTS'))

        other_testrun_fields = [
            'arch',
            'assignee',
            'build',
            'compose_id',
            'description',
            'fips',
            'logs',
            'planned_in',
            'platform',
            'pool_team',
            'sample_image',
        ]

        testsuites_properties: dict[str, Optional[str]] = {}

        for tr_field in other_testrun_fields:
            param = self.get(tr_field, os.getenv(f'TMT_PLUGIN_REPORT_POLARION_{tr_field.upper()}'))
            # TODO: remove the os.getenv when envvars in click work with steps in plans as well
            # as with steps on cmdline
            if param:
                testsuites_properties[f"polarion-custom-{tr_field.replace('_', '')}"] = param

        if use_facts:
            guests = self.step.plan.provision.ready_guests
            try:
                testsuites_properties['polarion-custom-hostname'] = guests[0].primary_address
                testsuites_properties['polarion-custom-arch'] = guests[0].facts.arch
            except IndexError as error:
                raise tmt.utils.ReportError(
                    'Failed to retrieve facts from the guest environment. '
                    'You can use a `--no-use-facts` argument to disable '
                    'this behavior.'
                ) from error

        if template:
            testsuites_properties['polarion-testrun-template-id'] = template

        logs = os.getenv('TMT_REPORT_ARTIFACTS_URL')
        if logs and 'polarion-custom-logs' not in testsuites_properties:
            testsuites_properties['polarion-custom-logs'] = logs

        project_span_ids: list[str] = []

        results_context = ResultsContext(self.step.plan.execute.results())

        for result in results_context:
            if not result.ids or not any(result.ids.values()):
                self.warn(
                    f"Test Case '{result.name}' is not exported to Polarion, "
                    "please run 'tmt tests export --how polarion' on it."
                )
                continue

            work_item_id, test_project_id = find_polarion_case_ids(result.ids)

            if work_item_id is None or test_project_id is None:
                self.warn(f"Test case '{result.name}' missing or not found in Polarion.")
                continue

            if test_project_id not in project_span_ids:
                project_span_ids.append(test_project_id)

            testcase_properties = {
                'polarion-testcase-id': work_item_id,
                'polarion-testcase-project-id': test_project_id,
            }

            # ignore[assignment]: mypy does not support different types for property getter and
            # setter. The assignment is correct, but mypy cannot tell.
            # See https://github.com/python/mypy/issues/3004 for getter/setter discussions
            result.properties = testcase_properties  # type: ignore[assignment]

        assert self.workdir is not None

        testsuites_properties.update(
            {
                'polarion-project-id': project_id,
                'polarion-user-id': PolarionWorkItem._session.user_id,
                'polarion-testrun-title': title,
                'polarion-project-span-ids': ','.join([project_id, *project_span_ids]),
            }
        )

        # Add deployment mode if provided as a context variable
        deployment_mode = self.step.plan._fmf_context.get('deployment-mode', [])
        if deployment_mode:
            testsuites_properties.update({'polarion-custom-deploymentMode': deployment_mode[0]})

        # ignore[assignment]: mypy does not support different types for property getter
        # and setter. The assignment is correct, but mypy cannot tell.
        # See https://github.com/python/mypy/issues/3004 for getter/setter discussions
        results_context.properties = testsuites_properties  # type: ignore[assignment]

        xml_data = make_junit_xml(
            phase=self,
            flavor='polarion',
            prettify=self.data.prettify,
            include_output_log=self.data.include_output_log,
            results_context=results_context,
        )

        f_path = self.data.file or self.workdir / DEFAULT_FILENAME

        try:
            f_path.write_text(xml_data)
        except Exception as error:
            raise tmt.utils.ReportError(f"Failed to write the output '{f_path}'.") from error

        if upload:
            server_url = str(PolarionWorkItem._session._server.url)
            polarion_import_url = (
                f'{server_url}{"" if server_url.endswith("/") else "/"}import/xunit'
            )
            auth = (PolarionWorkItem._session.user_id, PolarionWorkItem._session.password)

            response = post(
                polarion_import_url,
                auth=auth,
                files={
                    'file': ('xunit.xml', xml_data),
                },
                timeout=10,
            )
            self.info(f'Response code is {response.status_code} with text: {response.text}')
        else:
            self.info('Polarion upload can be done manually using command:')
            self.info(
                'curl -k -u <USER>:<PASSWORD> -X POST -F file=@<XUNIT_XML_FILE_PATH> '
                '<POLARION_URL>/polarion/import/xunit'
            )
        self.info('xUnit file saved at', f_path, 'yellow')



"""
Test Metadata Utilities
"""

import functools
import os
import re
import subprocess
import urllib.parse
from re import Pattern
from typing import TYPE_CHECKING, Optional

import tmt.log
import tmt.utils
from tmt.container import container
from tmt.utils import (
    Command,
    CommandOutput,
    Common,
    Environment,
    GeneralError,
    GitUrlError,
    MetadataError,
    Path,
    RunError,
)

if TYPE_CHECKING:
    import tmt.base


@container
class GitInfo:
    """
    Data container for commonly queried git data.
    """

    #: Path to the git root.
    git_root: Path

    #: Most human-readable git ref.
    ref: str

    #: Git remote linked to the current git ref.
    remote: str

    #: Default branch of the remote.
    default_branch: Optional[str]

    #: Public url of the remote.
    url: Optional[str]

    @classmethod
    @functools.cache
    def from_fmf_root(cls, *, fmf_root: Path, logger: tmt.log.Logger) -> Optional["GitInfo"]:
        """
        Get the current git info of an fmf tree.

        :param fmf_root: Root path of the fmf tree
        :param logger: Current tmt logger
        :return: Git info container or ``None`` if git metadata could not be resolved
        """

        def run(command: Command) -> str:
            """
            Run command, return output.
            We don't need the stderr here, but we need exit status.
            """

            result = command.run(cwd=fmf_root, logger=logger)
            if result.stdout is None:
                return ""
            return result.stdout.strip()

        # Prepare url (for now handle just the most common schemas)
        try:
            # Check if we are a git repo
            run(Command("git", "rev-parse", "--is-inside-work-tree"))

            # Initialize common git facts
            # Get some basic git references for HEAD
            all_refs = run(Command("git", "for-each-ref", "--format=%(refname)", "--points-at=@"))
            logger.debug("git all_refs", all_refs, level=3)
            # curr_ref is either HEAD or fully-qualified (branch) reference
            curr_ref = run(Command("git", "rev-parse", "--symbolic-full-name", "@"))
            logger.debug("git initial curr_ref", curr_ref, level=3)
            # Get the top-level git_root
            _git_root = git_root(fmf_root=fmf_root, logger=logger)
            assert _git_root is not None  # narrow type
        except RunError:
            # Not a git repo, everything should be pointing to None at this point
            return None

        if curr_ref != "HEAD":
            # The reference is fully qualified -> we are on a branch
            # Get the short name
            branch = run(Command("git", "for-each-ref", "--format=%(refname:short)", curr_ref))
            ref = branch
        else:
            # Not on a branch, check if we are on a tag or just a refs
            try:
                tags = run(Command("git", "describe", "--tags"))
                logger.debug("git tags", tags, level=3)
                # Is it possible to find which tag was used to checkout?
                # Now we just assume the first tag is the one we want
                tag_used = tags.splitlines()[0]
                logger.debug("Using tag", tag_used, level=3)
                # Point curr_ref to the fully-qualified ref
                curr_ref = f"refs/tags/{tag_used}"
                ref = tag_used
            except RunError:
                # We are not on a tag, just use the first available reference
                curr_ref = all_refs.splitlines()[0] if all_refs else curr_ref
                # Point the ref to the commit
                commit = run(Command("git", "rev-parse", curr_ref))
                logger.debug("Using commit", commit, level=3)
                ref = commit

        logger.debug("curr_ref used", curr_ref, level=3)
        remote_name = run(
            Command("git", "for-each-ref", "--format=%(upstream:remotename)", curr_ref)
        )
        if not remote_name:
            # If no specific upstream is defined, default to `origin`
            remote_name = "origin"
        try:
            remote = run(Command("git", "config", "--get", f"remote.{remote_name}.url"))
            url = public_git_url(remote)
            _default_branch = default_branch(repository=_git_root, remote=remote, logger=logger)
        except RunError:
            url = None
            _default_branch = None

        return GitInfo(
            git_root=_git_root,
            ref=ref,
            remote=remote_name,
            url=url,
            default_branch=_default_branch,
        )


# Avoid multiple subprocess calls for the same url
@functools.cache
def check_git_url(url: str, logger: tmt.log.Logger) -> str:
    """
    Check that a remote git url is accessible
    """

    try:
        logger.debug(f"Check git url '{url}'.")
        subprocess.check_call(
            ["git", "ls-remote", "--heads", url],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            env={"GIT_ASKPASS": "echo", "GIT_TERMINAL_PROMPT": "0"},
        )
        return url
    except subprocess.CalledProcessError:
        raise GitUrlError(f"Unable to contact remote git via '{url}'.")


PUBLIC_GIT_URL_PATTERNS: list[tuple[str, str]] = [
    # Gitlab on private namespace is synced to pkgs.devel.redhat.com
    # old: https://gitlab.com/redhat/rhel/tests/bash
    # old: git@gitlab.com:redhat/rhel/tests/bash
    # new: https://pkgs.devel.redhat.com/git/tests/bash
    (
        r'(?:git@|https://)gitlab.com[:/]redhat/rhel(/.+)',
        r'https://pkgs.devel.redhat.com/git\1',
    ),
    # GitHub, GitLab
    # old: git@github.com:teemtee/tmt.git
    # new: https://github.com/teemtee/tmt.git
    (
        r'git@(.*):(.*)',
        r'https://\1/\2',
    ),
    # RHEL packages
    # old: git+ssh://psplicha@pkgs.devel.redhat.com/tests/bash
    # old: ssh://psplicha@pkgs.devel.redhat.com/tests/bash
    # old: ssh://pkgs.devel.redhat.com/tests/bash
    # new: https://pkgs.devel.redhat.com/git/tests/bash
    (
        r'(git\+)?ssh://(\w+@)?(pkgs\.devel\.redhat\.com)/(.*)',
        r'https://\3/git/\4',
    ),
    # Fedora packages, Pagure
    # old: git+ssh://psss@pkgs.fedoraproject.org/tests/shell
    # old: ssh://psss@pkgs.fedoraproject.org/tests/shell
    # new: https://pkgs.fedoraproject.org/tests/shell
    (
        r'(git\+)?ssh://(\w+@)?([^/]*)/(.*)',
        r'https://\3/\4',
    ),
]


def public_git_url(url: str) -> str:
    """
    Convert a git url into a public format.

    :param url: an URL to convert.
    :returns: URL that is publicly accessible without authentication,
        or the original URL if no applicable conversion was found.
    """

    return rewrite_git_url(url, PUBLIC_GIT_URL_PATTERNS)


def rewrite_git_url(url: str, patterns: list[tuple[str, str]]) -> str:
    """
    Rewrite git url based on supplied patterns

    :param url: an URL to modify
    :param patterns: List of patterns to try in order
    :returns: Modified url or the original one if no pattern was be applied.
    """

    for pattern, replacement in patterns:
        public_url = re.sub(pattern, replacement, url)

        # If we got different string, `pattern` matched the URL and
        # `replacement` made its changes - we got our hit!
        if public_url != url:
            return public_url

    # Otherwise return unmodified
    return url


# Environment variable prefixes
INJECT_CREDENTIALS_URL_PREFIX = 'TMT_GIT_CREDENTIALS_URL_'
INJECT_CREDENTIALS_VALUE_PREFIX = 'TMT_GIT_CREDENTIALS_VALUE_'


def inject_auth_git_url(url: str) -> str:
    """
    Inject username or token to the git url

    :param url: original git repo url
    :returns: URL with injected authentication based on pattern from the environment
        or unmodified URL
    """

    # Try all environment variables sorted by their name
    for name, value in sorted(os.environ.items(), key=lambda x: x[0]):
        # First one which matches url is taken into the account
        if name.startswith(INJECT_CREDENTIALS_URL_PREFIX) and re.search(value, url):
            unique_suffix = name[len(INJECT_CREDENTIALS_URL_PREFIX) :]
            variable_with_value = f'{INJECT_CREDENTIALS_VALUE_PREFIX}{unique_suffix}'
            # Get credentials value
            try:
                creds = os.environ[variable_with_value]
            except KeyError:
                raise GitUrlError(
                    f'Missing "{variable_with_value}" variable with credentials for "{url}"'
                )
            # Return original url if credentials is an empty value
            if not creds:
                return url
            # Finally inject credentials into the url and return it
            return re.sub(r'([^/]+://)([^/]+)', rf'\1{creds}@\2', url)
    # Otherwise return unmodified
    return url


CLONABLE_GIT_URL_PATTERNS: list[tuple[str, str]] = [
    # git:// protocol is not possible for r/o access
    # old: git://pkgs.devel.redhat.com/tests/bash
    # new: https://pkgs.devel.redhat.com/git/tests/bash
    (
        r'git://(pkgs\.devel\.redhat\.com)/(.*)',
        r'https://\1/git/\2',
    ),
]


def clonable_git_url(url: str) -> str:
    """
    Modify the git repo url so it can be cloned
    """

    url = rewrite_git_url(url, CLONABLE_GIT_URL_PATTERNS)
    return inject_auth_git_url(url)


def web_git_url(url: str, ref: str, path: Optional[Path] = None) -> str:
    """
    Convert a public git url into a clickable web url format

    Compose a clickable link from git url, ref and path to file
    for the most common git servers.
    """

    if path:
        path = Path(urllib.parse.quote_plus(str(path), safe="/"))

    # Special handling for pkgs.devel (ref at the end)
    if 'pkgs.devel' in url:
        url = url.replace('git://', 'https://').replace('.com', '.com/cgit')
        url += '/tree'
        if path:
            url += str(path)
        url += f'?h={ref}'
        return url

    # GitHub & GitLab
    if any(server in url for server in ['github', 'gitlab']):
        url = url.replace('.git', '').rstrip('/')
        url += f'/tree/{ref}'

    if path:
        url += str(path)

    return url


def git_hash(*, directory: Path, logger: tmt.log.Logger) -> Optional[str]:
    """
    Return short hash of current HEAD in the git repo in directory.

    :param directory: path to a local git repository.
    :param logger: used for logging.
    :returns: short hash as string
    """

    cmd = Command("git", "rev-parse", "--short", "HEAD")
    result = cmd.run(cwd=directory, logger=logger)

    if result.stdout is None:
        raise RunError(
            message="No output from 'git' when looking for the hash of HEAD.",
            command=cmd,
            returncode=0,
            stderr=result.stderr,
        )

    return result.stdout.strip()


def git_root(*, fmf_root: Path, logger: tmt.log.Logger) -> Optional[Path]:
    """
    Find a path to the root of git repository containing an fmf root.

    :param fmf_root: path to an fmf root that is supposedly in a git repository.
    :param logger: used for logging.
    :returns: path to the git repository root, if fmf root lies in one,
        or ``None``.
    """

    try:
        result = Command("git", "rev-parse", "--show-toplevel").run(cwd=fmf_root, logger=logger)

        if result.stdout is None:
            return None

        return Path(result.stdout.strip())

    except RunError:
        # Always return an empty string in case 'git' command is run in a non-git repo
        return None


def git_add(*, path: Path, logger: tmt.log.Logger) -> None:
    """
    Add path to the git index.

    :param path: path to add to the git index.
    :param logger: used for logging.
    """

    path = path.resolve()

    try:
        Command("git", "add", path).run(cwd=path.parent, logger=logger)

    except RunError as error:
        raise GeneralError(f"Failed to add path '{path}' to git index.") from error


def git_ignore(*, root: Path, logger: tmt.log.Logger) -> list[Path]:
    """
    Collect effective paths ignored by git.

    :param root: path to the root of git repository.
    :param logger: used for logging.
    :returns: list of actual paths that would be ignored by git based on
        its ``.gitignore`` files. If a whole directory is to be ignored,
        it is listed as a directory path, not listing its content.
    """

    output = Command(
        'git',
        'ls-files',
        # Consider standard git exclusion files
        '--exclude-standard',
        # List untracked files matching exclusion patterns
        '-oi',
        # If a whole directory is to be ignored, list only its name with a trailing slash
        '--directory',
    ).run(cwd=root, logger=logger)

    return [Path(line.strip()) for line in output.stdout.splitlines()] if output.stdout else []


def default_branch(
    *,
    repository: Path,
    remote: str = 'origin',
    logger: tmt.log.Logger,
) -> Optional[str]:
    """
    Detect default branch from given local git repository
    """

    # Make sure '.git' is present and it is a file or a directory
    dot_git = repository / '.git'
    if not dot_git.exists():
        return None
    if not dot_git.is_file() and not dot_git.is_dir():
        return None

    # Detect the original repository path if worktree is provided
    if dot_git.is_file():
        try:
            result = Command("git", "rev-parse", "--path-format=absolute", "--git-common-dir").run(
                cwd=repository,
                logger=logger,
            )
        except RunError:
            return None
        if result.stdout is None:
            return None
        repository = Path(result.stdout.strip().replace("/.git", ""))

    # Make sure the '.git/refs/remotes/{remote}' directory is present
    git_remotes_dir = repository / f'.git/refs/remotes/{remote}'
    if not git_remotes_dir.exists():
        return None

    # Make sure the HEAD reference is available
    head = git_remotes_dir / 'HEAD'
    if not head.exists():
        try:
            Command('git', 'remote', 'set-head', f'{remote}', '--auto').run(
                cwd=repository,
                logger=logger,
            )
        except BaseException:
            return None

    # The ref format is 'ref: refs/remotes/origin/main'
    return head.read_text().strip().split('/')[-1]


def validate_git_status(test: 'tmt.base.Test') -> tuple[bool, str]:
    """
    Validate that test has current metadata on fmf_id

    Return a tuple (boolean, message) as the result of validation.

    Checks that sources:
    - all local changes are committed
    - up to date on remote repository
    - .fmf/version marking fmf root is committed as well

    When all checks pass returns (True, '').
    """

    # There has to be an fmf tree root defined
    if not test.fmf_root:
        raise MetadataError(f"Test '{test.name}' does not have fmf root defined.")

    sources = [
        *test.fmf_sources,
        test.fmf_root / '.fmf' / 'version',
    ]

    # Use tmt's run instead of subprocess.run
    run = Common(logger=test._logger).run

    # Narrow the type - this should be true, we might need to update
    # this assert once fmf is properly annotated
    assert isinstance(test.node.root, str)
    cwd = Path(test.node.root)

    # Check for not committed metadata changes
    cmd = Command(
        'git',
        'status',
        '--porcelain',
        '--',
        *[str(source) for source in sources],
    )
    try:
        result = run(cmd, cwd=cwd, join=True)
    except RunError as error:
        return (False, f"Failed to run git status: {error.stdout}")

    not_committed: list[str] = []
    assert result.stdout is not None
    for line in result.stdout.split('\n'):
        if line:
            # XY PATH or XY ORIG -> PATH. XY and PATH are separated by space
            not_committed.append(line[3:])

    if not_committed:
        return (False, "Uncommitted changes in " + " ".join(not_committed))

    # Check for not pushed changes
    cmd = Command("git", "rev-parse", "--abbrev-ref", "--symbolic-full-name", "@{u}")
    try:
        result = run(cmd, cwd=cwd)
    except RunError as error:
        return (False, f'Failed to get remote branch, error raised: "{error.stderr}"')

    assert result.stdout is not None
    remote_ref = result.stdout.strip()

    cmd = Command(
        'git',
        'diff',
        f'HEAD..{remote_ref}',
        '--name-status',
        '--',
        *[str(source) for source in sources],
    )
    try:
        result = run(cmd, cwd=cwd)
    except RunError as error:
        return (False, f'Failed to diff against remote branch, error raised: "{error.stderr}"')

    not_pushed: list[str] = []
    assert result.stdout is not None
    for line in result.stdout.split('\n'):
        if line:
            _, path = line.strip().split('\t', maxsplit=2)
            not_pushed.append(path)
    if not_pushed:
        return (False, "Not pushed changes in " + " ".join(not_pushed))

    return (True, '')


class DistGitHandler:
    """
    Common functionality for DistGit handlers
    """

    sources_file_name = 'sources'
    uri = "/rpms/{name}/{filename}/{hashtype}/{hash}/{filename}"

    usage_name: str  # Name to use for dist-git-type
    re_source: Pattern[str]
    # https://www.gnu.org/software/tar/manual/tar.html#auto_002dcompress
    re_supported_extensions: Pattern[str] = re.compile(
        r'\.((tar\.(gz|Z|bz2|lz|lzma|lzo|xz|zst))|tgz|taz|taZ|tz2|tbz2|tbz|tlz|tzst)$'
    )
    lookaside_server: str
    remote_substring: Pattern[str]

    def url_and_name(self, cwd: Optional[Path] = None) -> list[tuple[str, str]]:
        """
        Return list of urls and basenames of the used source

        The 'cwd' parameter has to be a DistGit directory.
        """

        cwd = cwd or Path.cwd()
        # Assumes <package>.spec
        globbed = list(cwd.glob('*.spec'))
        if len(globbed) != 1:
            raise GeneralError(f"No .spec file is present in '{cwd}'.")
        package = globbed[0].stem
        ret_values: list[tuple[str, str]] = []
        try:
            for line in (cwd / self.sources_file_name).splitlines():
                match = self.re_source.match(line)
                if match is None:
                    raise GeneralError(
                        f"Couldn't match '{self.sources_file_name}' "
                        f"content with '{self.re_source.pattern}'."
                    )
                used_hash, source_name, hash_value = match.groups()
                ret_values.append(
                    (
                        self.lookaside_server
                        + self.uri.format(
                            name=package,
                            filename=source_name,
                            hash=hash_value,
                            hashtype=used_hash.lower(),
                        ),
                        source_name,
                    )
                )
        except Exception as error:
            raise GeneralError(f"Couldn't read '{self.sources_file_name}' file.") from error
        if not ret_values:
            raise GeneralError("No sources found in '{self.sources_file_name}' file.")
        return ret_values

    def its_me(self, remotes: list[str]) -> bool:
        """
        True if self can work with remotes
        """

        return any(self.remote_substring.search(item) for item in remotes)


class FedoraDistGit(DistGitHandler):
    """
    Fedora Handler
    """

    usage_name = "fedora"
    re_source = re.compile(r"^(\w+) \(([^)]+)\) = ([0-9a-fA-F]+)$")
    lookaside_server = "https://src.fedoraproject.org/repo/pkgs"
    remote_substring = re.compile(r'fedoraproject\.org')


class CentOSDistGit(DistGitHandler):
    """
    CentOS Handler
    """

    usage_name = "centos"
    re_source = re.compile(r"^(\w+) \(([^)]+)\) = ([0-9a-fA-F]+)$")
    lookaside_server = "https://sources.stream.centos.org/sources"
    remote_substring = re.compile(r'redhat/centos')


class RedHatDistGit(DistGitHandler):
    """
    Red Hat Handler
    """

    usage_name = "redhat"
    re_source = re.compile(r"^(\w+) \(([^)]+)\) = ([0-9a-fA-F]+)$")
    # Location already public (standard-test-roles)
    lookaside_server = "http://pkgs.devel.redhat.com/repo"
    remote_substring = re.compile(r'redhat/rhel/|pkgs\.devel\.redhat\.com')


def get_distgit_handler(
    remotes: Optional[list[str]] = None,
    usage_name: Optional[str] = None,
) -> DistGitHandler:
    """
    Return the right DistGitHandler

    Pick the DistGitHandler class which understands specified
    remotes or by usage_name.
    """

    for candidate_class in DistGitHandler.__subclasses__():
        if usage_name is not None and usage_name == candidate_class.usage_name:
            return candidate_class()
        if remotes is not None:
            ret_val = candidate_class()
            if ret_val.its_me(remotes):
                return ret_val
    raise GeneralError(f"No known remote in '{remotes}'.")


def get_distgit_handler_names() -> list[str]:
    """
    All known distgit handlers
    """

    return [i.usage_name for i in DistGitHandler.__subclasses__()]


def distgit_download(
    *,
    distgit_dir: Path,
    target_dir: Path,
    handler_name: Optional[str] = None,
    caller: Optional['Common'] = None,
    logger: tmt.log.Logger,
) -> None:
    """
    Download sources to the target_dir

    distgit_dir is path to the DistGit repository
    """

    # Get the handler unless specified
    if handler_name is None:
        cmd = Command("git", "config", "--get-regexp", '^remote\\..*.url')
        output = cmd.run(cwd=distgit_dir, caller=caller, logger=logger)
        if output.stdout is None:
            raise tmt.utils.GeneralError("Missing remote origin url.")
        remotes = output.stdout.split('\n')
        handler = get_distgit_handler(remotes=remotes)
    else:
        handler = get_distgit_handler(usage_name=handler_name)

    for url, source_name in handler.url_and_name(distgit_dir):
        logger.debug(f"Download sources from '{url}'.")
        with tmt.utils.retry_session() as session:
            response = session.get(url)
        response.raise_for_status()
        target_dir.mkdir(exist_ok=True, parents=True)
        (target_dir / source_name).write_bytes(response.content)


def git_clone(
    *,
    url: str,
    destination: Path,
    shallow: bool = False,
    can_change: bool = True,
    env: Optional[Environment] = None,
    attempts: Optional[int] = None,
    interval: Optional[int] = None,
    timeout: Optional[int] = None,
    logger: tmt.log.Logger,
) -> CommandOutput:
    """
    Clone git repository from provided url to the destination directory

    :param url: Source URL of the git repository.
    :param destination: Full path to the destination directory.
    :param shallow: For ``shallow=True`` first try to clone repository
        using ``--depth=1`` option. If not successful clone repo with
        the whole history.
    :param can_change: URL can be modified with hardcoded rules. Use
        ``can_change=False`` to disable rewrite rules.
    :param env: Environment provided to the ``git clone`` process.
    :param attempts: Number of tries to call the function.
    :param interval: Amount of seconds to wait before a new try.
    :param timeout: Overall maximum time in seconds to clone the repo.
    :param logger: A Logger instance to be used for logging.
    :returns: Command output, bundled in a :py:class:`CommandOutput` tuple.
    """

    def clone_the_repo(
        url: str,
        destination: Path,
        shallow: bool = False,
        env: Optional[Environment] = None,
        timeout: Optional[int] = None,
    ) -> CommandOutput:
        """
        Clone the repo, handle history depth
        """

        depth = ['--depth=1'] if shallow else []
        return Command('git', 'clone', *depth, url, destination).run(
            cwd=Path('/'), env=env, timeout=timeout, logger=logger
        )

    from tmt.utils import GIT_CLONE_ATTEMPTS, GIT_CLONE_INTERVAL, GIT_CLONE_TIMEOUT

    timeout = timeout or GIT_CLONE_TIMEOUT
    attempts = attempts or GIT_CLONE_ATTEMPTS
    interval = interval or GIT_CLONE_INTERVAL

    # Update url only once
    if can_change:
        url = clonable_git_url(url)

    # Do an extra shallow clone first
    if shallow:
        try:
            return clone_the_repo(
                shallow=True,
                url=url,
                destination=destination,
                env=env,
                timeout=timeout,
            )
        except RunError:
            logger.debug(f"Shallow clone of '{url}' failed, let's try with the full history.")

    # Finish with whatever number attempts requested (deep)
    return tmt.utils.retry(
        func=clone_the_repo,
        attempts=attempts,
        interval=interval,
        label=f"git clone {url} {destination}",
        url=url,
        destination=destination,
        shallow=False,
        env=env,
        timeout=timeout,
        logger=logger,
    )



import pytest

import tmt
import tmt.base
import tmt.libraries
import tmt.utils


def test_basic(root_logger, source_dir, target_dir):
    """
    Test basic scenario for file requirement
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    tmt.libraries.library_factory(
        logger=root_logger,
        parent=parent,
        identifier=tmt.base.DependencyFile(type='file', pattern=['lib.*']),
        source_location=source_dir,
        target_location=target_dir,
    )
    assert target_dir.exists()
    assert target_dir.is_dir()
    target_content = list(target_dir.iterdir())
    assert target_dir / 'library' in target_content
    assert target_dir / 'lib_folder' in target_content


def test_full_copy(root_logger, source_dir, target_dir):
    """
    Test copying everything from the source directory
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    tmt.libraries.library_factory(
        logger=root_logger,
        parent=parent,
        identifier=tmt.base.DependencyFile(type='file', pattern=['/']),
        source_location=source_dir,
        target_location=target_dir,
    )
    assert (target_dir / 'tests/bz6/runtests.sh').exists()


def test_nothing_found(root_logger, source_dir, target_dir):
    """
    Test Library error is thrown when no files are found
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    with pytest.raises(tmt.utils.MetadataError):
        tmt.libraries.library_factory(
            logger=root_logger,
            parent=parent,
            identifier=tmt.base.DependencyFile(type='file', pattern=['/should/not/exist']),
            source_location=source_dir,
            target_location=target_dir,
        )



from typing import Any, Optional

import pytest

import tmt.log
import tmt.utils
from tmt.container import SerializableContainer, container, field
from tmt.utils import dataclass_normalize_field


def test_sanity():
    pass


def test_field_normalize_callback(root_logger: tmt.log.Logger) -> None:
    def _normalize_foo(key_address: str, raw_value: Any, logger: tmt.log.Logger) -> int:
        if raw_value is None:
            return None

        try:
            return int(raw_value)

        except ValueError as exc:
            raise tmt.utils.NormalizationError(
                key_address, raw_value, 'unset or an integer'
            ) from exc

    @container
    class DummyContainer(SerializableContainer):
        foo: Optional[int] = field(
            default=1,
            normalize=_normalize_foo,
        )

    # Initialize a data package
    data = DummyContainer()
    assert data.foo == 1

    dataclass_normalize_field(data, ':foo', 'foo', None, root_logger)
    assert data.foo is None

    dataclass_normalize_field(data, ':foo', 'foo', 2, root_logger)
    assert data.foo == 2

    dataclass_normalize_field(data, ':foo', 'foo', '3', root_logger)
    assert data.foo == 3

    with pytest.raises(
        tmt.utils.SpecificationError,
        match=r"Field ':foo' must be unset or an integer, 'str' found.",
    ):
        dataclass_normalize_field(data, ':foo', 'foo', 'will crash', root_logger)

    assert data.foo == 3


def test_field_custom_serialize():
    @container
    class DummyContainer(SerializableContainer):
        foo: list[str] = field(
            default_factory=list,
            serialize=lambda foo: ['serialized-into'],
            unserialize=lambda serialized_foo: ['unserialized-from'],
        )
        bar: str = field(default='should-never-change')

    # Initialize a data package
    data = DummyContainer()
    assert data.foo == []

    # When serialized, custom callback should override whatever was assigned to package keys
    serialized = data.to_serialized()
    assert serialized['foo'] == ['serialized-into']
    assert serialized['bar'] == 'should-never-change'

    # Yep, the custom callback should be stronger than the original value.
    data.foo = ['baz']
    serialized = data.to_serialized()
    assert serialized['foo'] == ['serialized-into']
    assert serialized['bar'] == 'should-never-change'

    # And the custom unserialize callback should also take an effect: value stored in
    # the package would be serializd into a different one by the custom serialize callback,
    # but then, when unserializing, the custom unserialize callback should lead to the key
    # having yet another value.
    data = DummyContainer()
    serialized = data.to_serialized()
    # We cannot compare `data` and output of `from_serialized()` because the unserialization
    # loads different value into the `foo` key, on purpose.
    data = DummyContainer.from_serialized(serialized)
    assert data.foo == ['unserialized-from']
    assert serialized['bar'] == 'should-never-change'



from typing import Any, Optional

import tmt.base
import tmt.export
import tmt.utils
from tmt.utils import Path


@tmt.base.FmfId.provides_export('template')
@tmt.base.Test.provides_export('template')
@tmt.base.Plan.provides_export('template')
@tmt.base.Story.provides_export('template')
class TemplateExporter(tmt.export.ExportPlugin):
    @classmethod
    def render_template(
        cls,
        *,
        template_filepath: Optional[Path] = None,
        default_template_filename: str,
        keys: Optional[list[str]] = None,
        **variables: Any,
    ) -> str:
        return tmt.utils.templates.render_template_file(
            template_filepath or tmt.export.TEMPLATES_DIRECTORY / default_template_filename,
            KEYS=keys,
            **variables,
        )

    @classmethod
    def export_fmfid_collection(
        cls,
        fmf_ids: list[tmt.base.FmfId],
        keys: Optional[list[str]] = None,
        template: Optional[Path] = None,
        **kwargs: Any,
    ) -> str:
        return '\n\n'.join(
            [
                cls.render_template(
                    template_filepath=template,
                    default_template_filename='default-fmfid.j2',
                    keys=keys,
                    FMF_ID=fmf_id,
                )
                for fmf_id in fmf_ids
            ]
        )

    @classmethod
    def export_test_collection(
        cls,
        tests: list[tmt.base.Test],
        keys: Optional[list[str]] = None,
        template: Optional[Path] = None,
        **kwargs: Any,
    ) -> str:
        return '\n\n'.join(
            [
                cls.render_template(
                    template_filepath=template,
                    default_template_filename='default-test.j2',
                    keys=keys,
                    TEST=test,
                )
                for test in tests
            ]
        )

    @classmethod
    def export_plan_collection(
        cls,
        plans: list[tmt.base.Plan],
        keys: Optional[list[str]] = None,
        template: Optional[Path] = None,
        **kwargs: Any,
    ) -> str:
        return '\n\n'.join(
            [
                cls.render_template(
                    template_filepath=template,
                    default_template_filename='default-plan.j2',
                    keys=keys,
                    PLAN=plan,
                )
                for plan in plans
            ]
        )

    @classmethod
    def export_story_collection(
        cls,
        stories: list[tmt.base.Story],
        keys: Optional[list[str]] = None,
        template: Optional[Path] = None,
        include_title: bool = True,
        **kwargs: Any,
    ) -> str:
        return '\n\n'.join(
            [
                cls.render_template(
                    template_filepath=template,
                    default_template_filename='default-story.j2',
                    keys=keys,
                    STORY=story,
                    INCLUDE_TITLE=include_title,
                )
                for story in stories
            ]
        )



import os
import pathlib
from collections.abc import Iterator
from typing import Optional, Union


class Path(pathlib.PosixPath):
    # Apparently, `pathlib`` does not offer `relpath` transition between
    # parallel trees, instead, a `ValueError`` is raised when `self` does not
    # lie under `other`. Overriding the original implementation with one based
    # on `os.path.relpath()`, which is more suited to tmt code base needs.
    #
    # ignore[override]: does not match the signature on purpose, our use is
    # slightly less generic that what pathlib supports, to be usable with
    # os.path.relpath.
    def relative_to(self, other: Union[str, "Path"]) -> "Path":  # type: ignore[override]
        return Path(os.path.relpath(self, other))

    # * `Path.is_relative_to()`` has been added in 3.9
    # https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.is_relative_to
    #
    # * The original implementation calls `relative_to()`, which we just made
    # to return a relative path for all possible inputs, even those the original
    # implementation considers to not be relative to each other. Therefore, we
    # need to override `is_relative_to()` even for other Python versions, to not
    # depend on `ValueError` raised by the original `relative_to()`.
    def is_relative_to(self, other: "Path") -> bool:  # type: ignore[override]
        # NOTE: the following is not perfect, but it should be enough for
        # what tmt needs to know about its paths.

        # Acquire the relative path from the one we're given and the other...
        relpath = os.path.relpath(str(self), str(other))

        # ... and if the relative path starts with `..`, it means we had to
        # walk *up* from `other` and descend to `path`, therefore `path` cannot
        # be a subtree of `other`.

        return not relpath.startswith("..")

    def unrooted(self) -> "Path":
        """
        Return the path as if it was not starting in file system root
        """

        if self.is_absolute():
            return self.relative_to("/")

        return self

    def append_text(
        self,
        data: str,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
    ) -> int:
        """
        Open the file pointed to in text mode, append data to it, and close the file
        """

        with self.open('a', encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    def splitlines(
        self,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        keepends: bool = False,
    ) -> Iterator[str]:
        """
        Yield decoded lines of the pointed-to file as a sequence of strings
        """

        yield from self.read_text(encoding=encoding, errors=errors).splitlines(keepends=keepends)



from unittest.mock import MagicMock, patch

import pytest

import tmt
from tmt.steps import Phase
from tmt.utils import GeneralError


class TestPhaseAssertFeelingSafe:
    def setup_method(self):
        self.mock_logger = MagicMock()
        self.phase = Phase(logger=self.mock_logger)

    @pytest.mark.parametrize(
        ("tmt_version", "deprecated_version", "expect_warn", "expect_exception"),
        [
            ('1.30', '1.38', True, False),  # warn for older version
            ('1.4.0.dev1595+ga35d7140.d20240806', '1.38', True, False),  # warn for older version
            ('1.40', '1.38', False, True),  # raise exception for newer version
            ('1.38', '1.38', False, True),  # raise exception for same version
        ],
        ids=(
            'warn for older version',
            'warn for older version with commit ID',
            'raise exception for newer version',
            'raise exception for same version',
        ),
    )
    def test_assert_feeling_safe(
        self, tmt_version, deprecated_version, expect_warn, expect_exception
    ):
        with patch.object(self.phase, 'warn') as mock_warn:
            tmt.__version__ = tmt_version

            if expect_exception:
                with pytest.raises(GeneralError):
                    self.phase.assert_feeling_safe(deprecated_version, 'Local provision plugin')
            else:
                self.phase.assert_feeling_safe(deprecated_version, 'Local provision plugin')

            assert mock_warn.called == expect_warn

    def test_assert_feeling_safe_feeling_safe(self):
        with (
            patch.object(Phase, 'is_feeling_safe', True),
            patch.object(self.phase, 'warn') as mock_warn,
        ):
            tmt.__version__ = '1.40'
            self.phase.assert_feeling_safe('1.38', 'Local provision plugin')

            # Check that warn is not called when feeling safe
            assert not mock_warn.called



import enum
import functools
from typing import TYPE_CHECKING, Any, Callable, Generic, Optional, TypedDict, TypeVar, cast

import tmt.log
import tmt.utils
from tmt.container import (
    SerializableContainer,
    SpecBasedContainer,
    container,
    field,
    key_to_option,
)
from tmt.plugins import PluginRegistry
from tmt.utils import NormalizeKeysMixin

if TYPE_CHECKING:
    import tmt.base
    from tmt.result import CheckResult
    from tmt.steps.execute import TestInvocation
    from tmt.steps.provision import Guest


#: A type variable representing a :py:class:`Check` instances.
CheckT = TypeVar('CheckT', bound='Check')

CheckPluginClass = type['CheckPlugin[Any]']

_CHECK_PLUGIN_REGISTRY: PluginRegistry[CheckPluginClass] = PluginRegistry('test.check')


def provides_check(check: str) -> Callable[[CheckPluginClass], CheckPluginClass]:
    """
    A decorator for registering test checks.

    Decorate a test check plugin class to register its checks.
    """

    def _provides_check(check_cls: CheckPluginClass) -> CheckPluginClass:
        _CHECK_PLUGIN_REGISTRY.register_plugin(
            plugin_id=check,
            plugin=check_cls,
            logger=tmt.log.Logger.get_bootstrap_logger(),
        )

        return check_cls

    return _provides_check


def find_plugin(name: str) -> 'CheckPluginClass':
    """
    Find a plugin by its name.

    :raises GeneralError: when the plugin does not exist.
    """

    plugin = _CHECK_PLUGIN_REGISTRY.get_plugin(name)

    if plugin is None:
        raise tmt.utils.GeneralError(f"Test check '{name}' was not found in check registry.")

    return plugin


# A "raw" test check as stored in fmf node data.
class _RawCheck(TypedDict):
    how: str
    enabled: bool
    result: str


class CheckEvent(enum.Enum):
    """
    Events in test runtime when a check can be executed
    """

    BEFORE_TEST = 'before-test'
    AFTER_TEST = 'after-test'

    @classmethod
    def from_spec(cls, spec: str) -> 'CheckEvent':
        try:
            return CheckEvent(spec)
        except ValueError:
            raise tmt.utils.SpecificationError(f"Invalid test check event '{spec}'.")


class CheckResultInterpret(enum.Enum):
    INFO = 'info'
    RESPECT = 'respect'
    XFAIL = 'xfail'

    @classmethod
    def from_spec(cls, spec: str) -> 'CheckResultInterpret':
        try:
            return CheckResultInterpret(spec)
        except ValueError as err:
            raise ValueError(f"Invalid check result interpretation '{spec}'.") from err

    def to_spec(self) -> str:
        return self.value


@container
class Check(
    SpecBasedContainer[_RawCheck, _RawCheck],
    SerializableContainer,
    NormalizeKeysMixin,
):
    """
    Represents a single check from test's ``check`` field.

    Serves as a link between raw fmf/CLI specification and an actual
    check implementation/plugin.
    """

    how: str
    enabled: bool = field(
        default=True,
        is_flag=True,
        help='Whether the check is enabled or not.',
    )
    result: CheckResultInterpret = field(
        default=CheckResultInterpret.RESPECT,
        help='How to interpret the check result.',
        serialize=lambda result: result.value,
        unserialize=CheckResultInterpret.from_spec,
        choices=[value.value for value in CheckResultInterpret.__members__.values()],
    )

    @functools.cached_property
    def plugin(self) -> 'CheckPluginClass':
        return find_plugin(self.how)

    # ignore[override]: expected, we need to accept one extra parameter, `logger`.
    @classmethod
    def from_spec(  # type: ignore[override]
        cls,
        raw_data: _RawCheck,
        logger: tmt.log.Logger,
    ) -> 'Check':
        data = cls(how=raw_data['how'])
        data._load_keys(cast(dict[str, Any], raw_data), cls.__name__, logger)
        if raw_data.get("result"):
            data.result = CheckResultInterpret.from_spec(raw_data["result"])

        return data

    def to_spec(self) -> _RawCheck:
        spec = cast(_RawCheck, {key_to_option(key): value for key, value in self.items()})
        spec["result"] = self.result.to_spec()
        return spec

    def to_minimal_spec(self) -> _RawCheck:
        return self.to_spec()

    def go(
        self,
        *,
        event: CheckEvent,
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list['CheckResult']:
        """
        Run the check.

        :param event: when the check is running - before the test, after the test, etc.
        :param invocation: test invocation to which the check belongs to.
        :param environment: optional environment to set for the check.
        :param logger: logger to use for logging.
        :returns: list of results produced by checks.
        """

        # TODO: there's "skipped" outcome brewing, we should use it once
        # it lands
        if not self.enabled:
            return []

        if event == CheckEvent.BEFORE_TEST:
            return self.plugin.before_test(
                check=self,
                invocation=invocation,
                environment=environment,
                logger=logger,
            )

        if event == CheckEvent.AFTER_TEST:
            return self.plugin.after_test(
                check=self,
                invocation=invocation,
                environment=environment,
                logger=logger,
            )

        raise tmt.utils.GeneralError(f"Unsupported test check event '{event}'.")


class CheckPlugin(tmt.utils._CommonBase, Generic[CheckT]):
    """
    Base class for plugins providing extra checks before, during and after tests
    """

    _check_class: type[CheckT]

    # Keep this method around, to correctly support Python's method resolution order.
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def delegate(
        cls,
        *,
        raw_data: _RawCheck,
        logger: tmt.log.Logger,
    ) -> Check:
        """
        Create a check data instance for the plugin
        """

        return cast(CheckPlugin[CheckT], find_plugin(raw_data['how']))._check_class.from_spec(
            raw_data, logger
        )

    @classmethod
    def essential_requires(
        cls,
        guest: 'Guest',
        test: 'tmt.base.Test',
        logger: tmt.log.Logger,
    ) -> list['tmt.base.DependencySimple']:
        """
        Collect all essential requirements of the test check.

        Essential requirements of a check are necessary for the check to
        perform its basic functionality.

        :returns: a list of requirements.
        """

        return []

    @classmethod
    def before_test(
        cls,
        *,
        check: CheckT,
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list['CheckResult']:
        return []

    @classmethod
    def after_test(
        cls,
        *,
        check: CheckT,
        invocation: 'TestInvocation',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list['CheckResult']:
        return []


def normalize_test_check(
    key_address: str,
    raw_test_check: Any,
    logger: tmt.log.Logger,
) -> Check:
    """
    Normalize a single test check
    """

    if isinstance(raw_test_check, str):
        try:
            return CheckPlugin.delegate(
                raw_data={'how': raw_test_check, 'enabled': True, 'result': 'respect'},
                logger=logger,
            )

        except Exception as exc:
            raise tmt.utils.SpecificationError(
                f"Cannot instantiate check from '{key_address}'."
            ) from exc

    if isinstance(raw_test_check, dict):
        try:
            return CheckPlugin.delegate(
                raw_data=cast(_RawCheck, raw_test_check),
                logger=logger,
            )

        except Exception as exc:
            raise tmt.utils.SpecificationError(
                f"Cannot instantiate check from '{key_address}'."
            ) from exc

    raise tmt.utils.NormalizationError(key_address, raw_test_check, 'a string or a dictionary')


def normalize_test_checks(
    key_address: str,
    raw_checks: Any,
    logger: tmt.log.Logger,
) -> list[Check]:
    """
    Normalize (prepare/finish/test) checks
    """

    if raw_checks is None:
        return []

    if isinstance(raw_checks, str):
        return [normalize_test_check(key_address, raw_checks, logger)]

    if isinstance(raw_checks, dict):
        return [normalize_test_check(key_address, raw_checks, logger)]

    if isinstance(raw_checks, list):
        # ignore[redundant-cast]: mypy infers the type to be `list[Any]` while
        # pyright, not making assumptions about the type of items, settles for
        # `list[Unknown]`. The `cast()` helps pyright, but mypy complains.
        return [
            normalize_test_check(f'{key_address}[{i}]', raw_test_check, logger)
            for i, raw_test_check in enumerate(cast(list[Any], raw_checks))  # type: ignore[redundant-cast]
        ]

    raise tmt.utils.NormalizationError(
        key_address, raw_checks, 'a string, a dictionary, or a list of their combinations'
    )



import re
from typing import TYPE_CHECKING, Any, Optional, cast

import tmt
import tmt.log
import tmt.steps
import tmt.steps.prepare
import tmt.utils
from tmt.container import container, field
from tmt.package_managers import Package
from tmt.result import PhaseResult
from tmt.steps.prepare import PreparePlugin
from tmt.steps.prepare.install import PrepareInstallData
from tmt.steps.provision import Guest
from tmt.utils import Command, Path, ShellScript, uniq

if TYPE_CHECKING:
    import tmt.base
    import tmt.steps.discover

#: Application of patches by ``rpm-build`` command invoked by
#: :ref:`discover/fmf</plugins/discover/fmf>` plugin when
#: ``dist-git-source`` key is used.
PHASE_ORDER_PREPARE_DISTGIT_RPMBUILD = 60

PREPARE_WRAPPER_FILENAME = 'tmt-prepare-wrapper.sh'

FEDORA_BUILD_REQUIRES = [Package('@buildsys-build')]
RHEL_BUILD_REQUIRES = [
    Package('tar'),
    Package('gcc-c++'),
    Package('redhat-rpm-config'),
    Package('redhat-release'),
    Package('which'),
    Package('xz'),
    Package('sed'),
    Package('make'),
    Package('bzip2'),
    Package('gzip'),
    Package('gcc'),
    Package('coreutils'),
    Package('unzip'),
    Package('diffutils'),
    Package('cpio'),
    Package('bash'),
    Package('gawk'),
    Package('info'),
    Package('patch'),
    Package('util-linux'),
    Package('findutils'),
    Package('grep'),
]


def insert_to_prepare_step(
    discover_plugin: 'tmt.steps.discover.DiscoverPlugin[Any]',
    sourcedir: Path,
) -> None:
    """
    Single place to call when inserting PrepareDistGit from discover
    """

    prepare_step = discover_plugin.step.plan.prepare
    where = cast(tmt.steps.discover.DiscoverStepData, discover_plugin.data).where
    # Future install require
    data_require = PrepareInstallData(
        how='install',
        name='requires (dist-git)',
        summary='Install required packages of tests detected by dist-git',
        order=tmt.steps.PHASE_ORDER_PREPARE_INSTALL_REQUIRES,
        where=where,
        package=[],
    )
    future_requires: PreparePlugin[Any] = cast(
        PreparePlugin[Any], PreparePlugin.delegate(prepare_step, data=data_require)
    )
    prepare_step._phases.append(future_requires)

    # Future install recommend
    data_recommend = PrepareInstallData(
        how='install',
        name='recommends (dist-git)',
        summary='Install recommended packages of tests detected by dist-git',
        order=tmt.steps.PHASE_ORDER_PREPARE_INSTALL_RECOMMENDS,
        where=where,
        package=[],
        missing='skip',
    )
    future_recommends: PreparePlugin[Any] = cast(
        PreparePlugin[Any], PreparePlugin.delegate(prepare_step, data=data_recommend)
    )
    prepare_step._phases.append(future_recommends)

    prepare_step._phases.append(
        PrepareDistGit(
            step=prepare_step,
            data=DistGitData(
                where=where,
                source_dir=sourcedir,
                phase_name=discover_plugin.name,
                install_builddeps=discover_plugin.get('dist-git-install-builddeps'),
                require=discover_plugin.get('dist-git-require'),
                how='distgit',
                name="Prepare dist-git sources (buildrequires, patches, discovery...)",
            ),
            workdir=None,
            discover=discover_plugin,
            future_requires=future_requires,
            future_recommends=future_recommends,
            logger=discover_plugin._logger.descend(logger_name="extract-distgit", extra_shift=0),
        )
    )


@container
class DistGitData(tmt.steps.prepare.PrepareStepData):
    source_dir: Optional[Path] = field(
        default=None,
        option='--source-dir',
        normalize=tmt.utils.normalize_path,
        exporter=lambda value: str(value) if isinstance(value, Path) else None,
        help="Path to the source directory where ``rpmbuild -bp`` should happen.",
        internal=True,
    )
    phase_name: str = field(
        default_factory=str,
        option='--phase-name',
        help="Name of the discover step phase to inject tests to.",
        internal=True,
    )
    order: int = PHASE_ORDER_PREPARE_DISTGIT_RPMBUILD
    install_builddeps: bool = field(
        default=False,
        option="--install-builddeps",
        is_flag=True,
        help="Install package build dependencies",
    )
    require: list['tmt.base.DependencySimple'] = field(
        default_factory=list,
        option="--require",
        metavar='PACKAGE',
        multiple=True,
        help='Additional required package(s) to be present before sources are prepared.',
        # *simple* requirements only
        normalize=lambda key_address, value, logger: tmt.base.assert_simple_dependencies(
            tmt.base.normalize_require(key_address, value, logger),
            "'require' can be simple packages only",
            logger,
        ),
        serialize=lambda packages: [package.to_spec() for package in packages],
        unserialize=lambda serialized: [
            tmt.base.DependencySimple.from_spec(package) for package in serialized
        ],
    )


# @tmt.steps.provides_method('distgit') Hiding from the menu
class PrepareDistGit(tmt.steps.prepare.PreparePlugin[DistGitData]):
    """
    Companion to the discover-dist-git, place where ``rpmbuild -bp`` happens

    Step is responsible:
    1. Install required packages for the rpmbuild itself
    2. Detect and install build requires
    3. Patch sources (rpmbuild -bp)
    4. Move patched sources from buildroot into TMT_SOURCE_DIR
    5. Call function of discover plugin to discover tests from TMT_SOURCE_DIR
    """

    _data_class = DistGitData

    def __init__(
        self,
        *,
        discover: Optional['tmt.steps.discover.DiscoverPlugin[Any]'] = None,
        future_requires: Optional['tmt.steps.prepare.PreparePlugin[Any]'] = None,
        future_recommends: Optional['tmt.steps.prepare.PreparePlugin[Any]'] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.discover = discover
        self.future_requires = future_requires
        self.future_recommends = future_recommends

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Prepare the guests for building rpm sources
        """

        results = super().go(guest=guest, environment=environment, logger=logger)

        environment = environment or tmt.utils.Environment()

        # Packages required for this plugin to work and additional required packages
        explicit_requires = [Package(p) for p in self.get('require', [])] + [Package('rpm-build')]

        # Packages assumed to be present when building packages
        guest_distro = guest.facts.distro.lower() if guest.facts.distro else ""
        if "fedora" in guest_distro:
            explicit_requires += FEDORA_BUILD_REQUIRES
        elif "red hat" in guest_distro:
            explicit_requires += RHEL_BUILD_REQUIRES

        if self.get('install_builddeps'):
            # FIXME For now dnf only, ideally it will be capability of the package_manager...
            if "dnf" not in (guest.facts.package_manager or ''):
                raise tmt.utils.PrepareError("Cannot install build deps on system without dnf yet")
            explicit_requires += [Package('dnf-command(builddep)')]

        # Install required packages for rpm-build to work
        guest.package_manager.install(*explicit_requires)

        source_dir = self.data.source_dir
        assert source_dir

        try:
            spec_name = next(Path(source_dir).glob('*.spec')).name
        except StopIteration:
            raise tmt.utils.PrepareError(f"No '*.spec' file found in '{source_dir}'")

        content_before = set(source_dir.iterdir())

        dir_defines = [
            "--define",
            f'_sourcedir {source_dir}',
            "--define",
            f'_builddir {source_dir}',
            "--define",
            f'_srcrpmdir {source_dir}/SRPMS',
        ]

        if self.get('install_builddeps'):
            cmd = Command("rpmbuild", "-br", "--nodeps", spec_name, *dir_defines)
            try:
                stdout = guest.execute(command=cmd, cwd=source_dir).stdout
            except tmt.utils.RunError as error:
                # manpage says rpmbuild should return '11' for `-br --nodeps`
                # but it doesn't seem to be the case on f-39
                if error.returncode != 11:
                    raise tmt.utils.PrepareError("Unexpected return code of `rpmbuild -br` call.")
                stdout = error.stdout
            match = re.search(r'/SRPMS/(.*src.rpm)', stdout or '')
            if match:
                src_rpm_name = match.group(1)
            else:
                raise tmt.utils.PrepareError('No src.rpm file created by the `rpmbuild -br` call.')
            # Install build requires
            # Create the package manager command
            cmd, _ = guest.package_manager.engine.prepare_command()
            # Can't set 'cwd' as the check for its existence fails for local workdir
            cmd += Command("builddep", "-y", f"SRPMS/{src_rpm_name}")
            guest.execute(command=cmd, cwd=Path(source_dir))

        # Finally run the rpm-build -bp
        cmd = Command("rpmbuild", "-bp", spec_name, "--nodeps", *dir_defines)
        try:
            guest.execute(
                command=cmd,
                cwd=source_dir,
            )
        except tmt.utils.RunError as error:
            raise tmt.utils.PrepareError("Unable to 'rpmbuild -bp'.", causes=[error])

        # Workaround around new rpm behavior, https://github.com/teemtee/tmt/issues/2987
        # No hardcoded name, should keep working in the future
        cmd = Command(
            "rpmbuild",
            "-bc",
            "--short-circuit",
            "--nodeps",
            "--define",
            '__spec_build_pre echo tmt-get-builddir=%{_builddir}; exit 0',
            spec_name,
            *dir_defines,
        )
        outcome = guest.execute(command=cmd, cwd=source_dir).stdout or ''
        match = re.search(r'tmt-get-builddir=(.+)', outcome)
        builddir = Path(match.group(1)) if match else None

        # But if the %build is missing in spec (e.g. in our test) the previous output was empty
        if builddir is None:
            guest.execute(
                command=ShellScript(
                    "shopt -s dotglob; if test -e */SPECPARTS; "
                    "then mv ./*-build/* .; else true; fi"
                ),
                cwd=source_dir,
            )
        elif builddir.resolve() != source_dir.resolve():
            guest.execute(command=ShellScript(f"shopt -s dotglob; mv {builddir}/* {source_dir}"))
        else:
            self.debug("Builddir matches source_dir, no need to copy anything.")

        # Make sure to pull back sources ...
        # FIXME -- Do we need to? Can be lot of data...
        guest.pull(source_dir)

        # Mark which file/dirs were created after rpmbuild ran
        content_after = set(source_dir.iterdir())
        created_content = [Path(p.name) for p in content_after - content_before]

        # When discover is set let it rediscover tests
        if self.discover is not None:
            self.discover.post_dist_git(created_content)
            # FIXME needs refactor of Prepare, tmt.base etc...
            # doing quick & dirty injection of prepareinstalls
            for g in self.step.plan.provision.ready_guests:
                collected_requires: list[tmt.base.DependencySimple] = []
                collected_recommends: list[tmt.base.DependencySimple] = []

                for test_origin in self.step.plan.discover.tests(enabled=True):
                    test = test_origin.test

                    if not test.enabled_on_guest(g):
                        continue

                    collected_requires += tmt.base.assert_simple_dependencies(
                        test.require,
                        'After beakerlib processing, tests may have only simple requirements',
                        self._logger,
                    )

                    collected_recommends += tmt.base.assert_simple_dependencies(
                        test.recommend,
                        'After beakerlib processing, tests may have only simple requirements',
                        self._logger,
                    )

                    collected_requires += test.test_framework.get_requirements(test, self._logger)

                    for check in test.check:
                        collected_requires += check.plugin.essential_requires(
                            guest, test, self._logger
                        )
                # Inject additional install plugins - require
                if collected_requires and self.future_requires:
                    self.future_requires.data.package = uniq(collected_requires)
                # Inject additional install plugins - recommend
                if collected_recommends and self.future_recommends:
                    self.future_recommends.data.package = uniq(collected_recommends)

        return results



import functools
import re
import sys
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Callable, Optional, TypedDict, Union, cast, overload

from jinja2 import FileSystemLoader, select_autoescape

import tmt
import tmt.log
import tmt.result
import tmt.steps
import tmt.steps.report
import tmt.utils
from tmt.container import container, field
from tmt.result import ResultOutcome
from tmt.utils import Path
from tmt.utils.templates import default_template_environment, render_template_file

if TYPE_CHECKING:
    from tmt._compat.typing import TypeAlias

    XMLElement: TypeAlias = Any


DEFAULT_FILENAME = 'junit.xml'
DEFAULT_FLAVOR_NAME = 'default'
CUSTOM_FLAVOR_NAME = 'custom'

# Relative path to tmt junit template directory.
DEFAULT_TEMPLATE_DIR = Path('steps/report/junit/templates/')


@overload
def _duration_to_seconds_filter(duration: str) -> int:
    pass


@overload
def _duration_to_seconds_filter(duration: None) -> None:
    pass


def _duration_to_seconds_filter(duration: Optional[str]) -> Optional[int]:
    """
    Convert valid duration string in to seconds
    """

    if duration is None:
        return None
    try:
        h, m, s = duration.split(':')
        return int(h) * 3600 + int(m) * 60 + int(s)
    except Exception as error:
        raise tmt.utils.ReportError(f"Malformed duration '{duration}'.") from error


def _escape_control_chars_filter(func: Callable[[str], str]) -> Callable[[str], str]:
    """
    Wrap the escape filter function and escape ASCII chosen control characters
    """

    def wrapper(value: str) -> str:
        # Define unicode characters which are not allowed in the XML and need to be removed.
        # - https://www.w3.org/TR/REC-xml/#NT-Char
        # - https://github.com/kyrus/python-junit-xml/blob/4bd08a272f059998cedf9b7779f944d49eba13a6/junit_xml/__init__.py#L325
        illegal_chars = [
            (0x00, 0x08),
            (0x0B, 0x1F),
            (0x7F, 0x84),
            (0x86, 0x9F),
            (0xD800, 0xDFFF),
            (0xFDD0, 0xFDDF),
            (0xFFFE, 0xFFFF),
            (0x1FFFE, 0x1FFFF),
            (0x2FFFE, 0x2FFFF),
            (0x3FFFE, 0x3FFFF),
            (0x4FFFE, 0x4FFFF),
            (0x5FFFE, 0x5FFFF),
            (0x6FFFE, 0x6FFFF),
            (0x7FFFE, 0x7FFFF),
            (0x8FFFE, 0x8FFFF),
            (0x9FFFE, 0x9FFFF),
            (0xAFFFE, 0xAFFFF),
            (0xBFFFE, 0xBFFFF),
            (0xCFFFE, 0xCFFFF),
            (0xDFFFE, 0xDFFFF),
            (0xEFFFE, 0xEFFFF),
            (0xFFFFE, 0xFFFFF),
            (0x10FFFE, 0x10FFFF),
        ]

        illegal_ranges = [
            f"{chr(low)}-{chr(high)}" for low, high in illegal_chars if low < sys.maxunicode
        ]

        illegal_regex = re.compile("[{}]".format("".join(illegal_ranges)))
        escaped_value = illegal_regex.sub("", value)

        # It's important to call the parent `func` at the end of the wrapper, otherwise the
        # jinja autoescape doesn't work correctly (some chars like '&' are escaped two times).
        return func(escaped_value)

    return wrapper


class ImplementProperties:
    """
    Define a properties attribute.

    This class can be used to easily add properties attribute by inheriting it.
    """

    class PropertyDict(TypedDict):
        """
        Defines a property dict, which gets propagated into the final template properties.
        """

        name: str
        value: str

    def __init__(self) -> None:
        self._properties: dict[str, str] = {}

    @property
    def properties(self) -> list[PropertyDict]:
        return [{'name': k, 'value': v} for k, v in self._properties.items()]

    @properties.setter
    def properties(self, keyval: dict[str, Optional[str]]) -> None:
        self._properties = {k: v for k, v in keyval.items() if v is not None}


class ResultWrapper(ImplementProperties):
    """
    The context wrapper for :py:class:`tmt.Result`.

    Adds possibility to wrap the :py:class:`tmt.Result` and dynamically add more attributes which
    get available inside the template context.
    """

    def __init__(
        self,
        wrapped: Union[tmt.Result, tmt.result.SubResult],
        subresults_context_class: 'type[ResultsContext]',
    ) -> None:
        super().__init__()
        self._wrapped = wrapped
        self._subresults_context_class = subresults_context_class

    def __getattr__(self, name: str) -> Any:
        """
        Returns an attribute of a wrapped ``tmt.Result`` instance
        """

        return getattr(self._wrapped, name)

    @property
    def subresult(self) -> 'ResultsContext':
        """
        Override the ``tmt.Result.subresult`` and wrap all the ``tmt.result.SubResult`` instances
        into the ``ResultsContext``.
        """

        # `tmt.result.SubResult.subresult` is not defined, just raise the AttributeError to silent
        # the typing errors.
        if isinstance(self._wrapped, tmt.result.SubResult):
            raise AttributeError(
                f"'{self._wrapped.__class__.__name__} object has no attribute 'subresult'"
            )

        return self._subresults_context_class(self._wrapped.subresult)


class ResultsContext(ImplementProperties):
    """
    The results context for Jinja templates.

    A class which keeps the results context (especially the result summary) for JUnit template. It
    wraps all the :py:class:`tmt.Result` instances into the :py:class:`ResultWrapper`.
    """

    def __init__(self, results: Union[list[tmt.Result], list[tmt.result.SubResult]]) -> None:
        """
        Decorate/wrap all the ``Result`` and ``SubResult`` instances with more attributes
        """

        super().__init__()

        # Decorate all the tmt.Results with more attributes
        self._results: list[ResultWrapper] = [
            ResultWrapper(r, subresults_context_class=self.__class__) for r in results
        ]

    def __iter__(self) -> Iterator[ResultWrapper]:
        """
        Possibility to iterate over results by iterating an instance
        """

        return iter(self._results)

    def __len__(self) -> int:
        """
        Returns the number of results
        """

        return len(self._results)

    @functools.cached_property
    def passed(self) -> list[ResultWrapper]:
        """
        Returns results of passed tests
        """

        return [r for r in self._results if r.result == ResultOutcome.PASS]

    @functools.cached_property
    def skipped(self) -> list[ResultWrapper]:
        """
        Returns results of skipped tests
        """

        return [r for r in self._results if r.result in (ResultOutcome.SKIP, ResultOutcome.INFO)]

    @functools.cached_property
    def failed(self) -> list[ResultWrapper]:
        """
        Returns results of failed tests
        """

        return [r for r in self._results if r.result == ResultOutcome.FAIL]

    @functools.cached_property
    def errored(self) -> list[ResultWrapper]:
        """
        Returns results of tests with error/warn outcome
        """

        return [r for r in self._results if r.result in (ResultOutcome.ERROR, ResultOutcome.WARN)]

    @functools.cached_property
    def duration(self) -> int:
        """
        Returns the total duration of all tests in seconds
        """

        # cast: mypy does not understand the proxy-ness of `ResultWrapper`. `r.duration`
        # will exists, therefore adding a `cast` to convince mypy the list is pretty much
        # nothing but the list of results.
        return sum(
            _duration_to_seconds_filter(r.duration) or 0
            for r in cast(list[tmt.Result], self._results)
        )


def make_junit_xml(
    phase: tmt.steps.report.ReportPlugin[Any],
    flavor: str = DEFAULT_FLAVOR_NAME,
    template_path: Optional[Path] = None,
    include_output_log: bool = True,
    prettify: bool = True,
    results_context: Optional[ResultsContext] = None,
    **extra_variables: Any,
) -> str:
    """
    Create JUnit XML file and return the data.

    :param phase: instance of a :py:class:`tmt.steps.report.ReportPlugin`.
    :param flavor: name of a JUnit flavor to generate.
    :param template_path: if set, the provided template will be used instead of a pre-defined
        flavor template. In this case, the ``flavor`` must be set to ``custom`` value.
    :param include_output_log: if enabled, the ``<system-out>`` tags are included in the final
        template output.
    :param prettify: allows to control the XML pretty print.
    :param results_context: if set, the provided :py:class:`ResultsContext` is used in a template.
    :param extra_variables: if set, these variables get propagated into the Jinja template.
    """

    # Get the template context for tmt results
    results_context = results_context or ResultsContext(phase.step.plan.execute.results())

    # Prepare the template environment
    environment = default_template_environment()

    template_path = template_path or tmt.utils.resource_files(
        DEFAULT_TEMPLATE_DIR / Path(f'{flavor}.xml.j2')
    )

    # Use a FileSystemLoader for a non-custom flavor
    if flavor != CUSTOM_FLAVOR_NAME:
        environment.loader = FileSystemLoader(
            searchpath=tmt.utils.resource_files(DEFAULT_TEMPLATE_DIR)
        )

    def _read_log_filter(log: Path) -> str:
        """
        Read the contents of a given result log
        """

        if not log:
            return ''

        try:
            return str(phase.step.plan.execute.read(log))
        except tmt.utils.FileError:
            return ''

    environment.filters.update(
        {
            'read_log': _read_log_filter,
            'duration_to_seconds': _duration_to_seconds_filter,
            'failures': tmt.result.Result.failures,
            # Use a wrapper function to also _escape_control_chars.
            'e': _escape_control_chars_filter(environment.filters['e']),
        }
    )

    # Explicitly enable the autoescape because it's disabled by default by tmt.
    # See /teemtee/tmt/issues/2873 for more info.
    environment.autoescape = select_autoescape(enabled_extensions=('xml.j2'))

    xml_data = render_template_file(
        template_path,
        environment,
        RESULTS=results_context,
        PLAN=phase.step.plan,
        INCLUDE_OUTPUT_LOG=include_output_log,
        **extra_variables,
    )

    # Try to use lxml to check the flavor XML schema and prettify the final XML
    # output.
    try:
        from lxml import etree

    except ImportError:
        from tmt.utils.hints import print_hints

        print_hints('report/junit', logger=phase._logger)

        return xml_data

    xml_parser_kwargs: dict[str, Any] = {
        'remove_blank_text': prettify,
        'huge_tree': True,
        'schema': None,
    }

    # The schema check must be done only for a non-custom JUnit flavors
    if flavor != CUSTOM_FLAVOR_NAME:
        xsd_schema_path = Path(
            tmt.utils.resource_files(Path(f'steps/report/junit/schemas/{flavor}.xsd'))
        )

        schema_root: XMLElement = etree.XML(xsd_schema_path.read_bytes())
        xml_parser_kwargs['schema'] = etree.XMLSchema(schema_root)
    else:
        phase.warn(
            f"The '{CUSTOM_FLAVOR_NAME}' JUnit flavor is used, you are solely responsible "
            "for the validity of the XML schema."
        )

        if prettify:
            phase.warn(
                f"The pretty print is always disabled for '{CUSTOM_FLAVOR_NAME}' JUnit flavor."
            )

        xml_parser_kwargs['remove_blank_text'] = prettify = False

    xml_parser = etree.XMLParser(**xml_parser_kwargs)
    try:
        # S320: Parsing of untrusted data is known to be vulnerable to XML attacks.
        tree_root: XMLElement = etree.fromstring(xml_data, xml_parser)  # noqa: S320

    except etree.XMLSyntaxError as e:
        phase.warn(
            'The generated XML output is not a valid XML file or it is not valid against the '
            'XSD schema.'
        )

        if flavor != CUSTOM_FLAVOR_NAME:
            phase.warn('Please, report this problem to project maintainers.')

        for err in e.error_log:
            phase.warn(str(err))

        # Return the prettified XML without checking the XSD
        del xml_parser_kwargs['schema']

        xml_parser = etree.XMLParser(**xml_parser_kwargs)

        try:
            tree_root = etree.fromstring(xml_data, xml_parser)  # noqa: S320
        except etree.XMLSyntaxError as error:
            phase.verbose('rendered XML', xml_data, 'red')
            raise tmt.utils.ReportError(
                'The generated XML output is not a valid XML file. Use `--verbose` argument '
                'to show the output.'
            ) from error

    # Do not be fooled by the `encoding` parameter: even with `utf-8`,
    # `tostring()` will still return bytes. `unicode`, on the other
    # hand, would give us a string, but then, no XML declaration.
    # So, we get bytes, and we need to apply `decode()` on our own.
    xml_output: bytes = etree.tostring(
        tree_root,
        xml_declaration=True,
        pretty_print=prettify,
        # The 'utf-8' encoding must be used instead of 'unicode', otherwise
        # the XML declaration is not included in the output.
        encoding='utf-8',
    )

    return str(xml_output.decode('utf-8'))


@container
class ReportJUnitData(tmt.steps.report.ReportStepData):
    file: Optional[Path] = field(
        default=None,
        option='--file',
        metavar='PATH',
        help='Path to the file to store JUnit to.',
        normalize=tmt.utils.normalize_path,
    )

    flavor: str = field(
        default=DEFAULT_FLAVOR_NAME,
        option='--flavor',
        choices=[DEFAULT_FLAVOR_NAME, CUSTOM_FLAVOR_NAME, 'subresults'],
        help='Name of a JUnit flavor to generate.',
    )

    template_path: Optional[Path] = field(
        default=None,
        option='--template-path',
        metavar='TEMPLATE_PATH',
        help='Path to a custom template file to use for JUnit creation.',
        normalize=tmt.utils.normalize_path,
    )

    prettify: bool = field(
        default=True,
        option=('--prettify / --no-prettify'),
        is_flag=True,
        show_default=True,
        help=f"""
            Enable the XML pretty print for generated JUnit file. This option is always disabled
            for '{CUSTOM_FLAVOR_NAME}' template flavor.
            """,
    )

    include_output_log: bool = field(
        default=True,
        option=('--include-output-log / --no-include-output-log'),
        is_flag=True,
        show_default=True,
        help='Include full standard output in resulting xml file.',
    )


@tmt.steps.provides_method(
    'junit',
    installation_hint="""
        For neater JUnit XML and XML validation against the XSD, ``lxml`` package is required
        by the ``report/junit`` plugin.

        To quickly test ``lxml`` presence, you can try running ``python -c 'import lxml'``.

        * Users who installed tmt from system repositories should install ``tmt+report-junit``
          package.
        * Users who installed tmt from PyPI should install ``tmt[report-junit]`` extra.
    """,
)
class ReportJUnit(tmt.steps.report.ReportPlugin[ReportJUnitData]):
    """
    Save test results in chosen JUnit flavor format.

    When flavor is set to custom, the ``template-path`` with a path to a custom template must be
    provided.

    When ``file`` is not specified, output is written into a file named ``junit.xml`` located in
    the current workdir.

    .. code-block:: yaml

        # Enable junit report from the command line
        tmt run --all report --how junit
        tmt run --all report --how junit --file test.xml

    .. code-block:: yaml

        # Use junit as the default report for given plan
        report:
            how: junit
            file: test.xml
    """

    _data_class = ReportJUnitData

    def check_options(self) -> None:
        """
        Check the module options
        """

        if self.data.flavor == 'custom' and not self.data.template_path:
            raise tmt.utils.ReportError(
                "The 'custom' flavor requires the '--template-path' argument."
            )

        if self.data.flavor != 'custom' and self.data.template_path:
            raise tmt.utils.ReportError(
                "The '--template-path' can be used only with '--flavor=custom'."
            )

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        members = super()._preserved_workdir_members

        if self.data.file is None:
            members = {*members, DEFAULT_FILENAME}

        return members

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Read executed tests and write junit
        """

        super().go(logger=logger)

        self.check_options()

        assert self.workdir is not None
        f_path = self.data.file or self.workdir / DEFAULT_FILENAME

        xml_data = make_junit_xml(
            phase=self,
            flavor=self.data.flavor,
            template_path=self.data.template_path,
            include_output_log=self.data.include_output_log,
            prettify=self.data.prettify,
        )
        try:
            f_path.write_text(xml_data)
            self.info('output', f_path, 'yellow')
        except Exception as error:
            raise tmt.utils.ReportError(f"Failed to write the output '{f_path}'.") from error



import textwrap

import pytest

import tmt.utils
from tests.unit.test_hardware import FULL_HARDWARE_REQUIREMENTS, OR_HARDWARE_REQUIREMENTS
from tmt.hardware import (
    Hardware,
    Operator,
    _parse_cpu,
    _parse_disk,
    _parse_hostname,
    _parse_iommu,
    _parse_location,
    _parse_memory,
    _parse_system,
    _parse_tpm,
    _parse_virtualization,
    _parse_zcrypt,
)
from tmt.log import Logger
from tmt.steps.provision.mrack import (
    _CONSTRAINT_TRANSFORMERS,
    constraint_to_beaker_filter,
    operator_to_beaker_op,
)


@pytest.mark.parametrize(
    ('operator', 'value', 'expected'),
    [
        (Operator.EQ, 'foo', ('==', 'foo', False)),
        (Operator.NEQ, 'foo', ('!=', 'foo', False)),
        (Operator.GT, 'foo', ('>', 'foo', False)),
        (Operator.GTE, 'foo', ('>=', 'foo', False)),
        (Operator.LT, 'foo', ('<', 'foo', False)),
        (Operator.LTE, 'foo', ('<=', 'foo', False)),
        (Operator.MATCH, 'f.+o.*o', ('like', 'f%o%o', False)),
        (Operator.NOTMATCH, 'f.+o.*o', ('like', 'f%o%o', True)),
    ],
)
def test_operator_to_beaker_op(
    operator: Operator, value: str, expected: tuple[str, str, bool]
) -> None:
    assert operator_to_beaker_op(operator, value) == expected


def test_maximal_constraint(root_logger: Logger) -> None:
    hw = Hardware.from_spec(tmt.utils.yaml_to_dict(textwrap.dedent(FULL_HARDWARE_REQUIREMENTS)))
    assert hw.constraint is not None

    result = constraint_to_beaker_filter(hw.constraint, root_logger)
    assert result.to_mrack() == {
        'and': [
            {
                'pool': {
                    '_op': '!=',
                    '_value': 'foo.*',
                },
            },
            {},
            {'and': [{}, {}]},
            {
                'and': [
                    {
                        'cpu': {
                            'processors': {'_op': '>', '_value': '8'},
                        },
                    },
                    {},
                    {
                        'cpu': {
                            'cores': {'_op': '==', '_value': '2'},
                        },
                    },
                    {},
                    {},
                    {},
                    {'cpu': {'model': {'_op': '==', '_value': '62'}}},
                    {'cpu': {'family': {'_op': '<', '_value': '6'}}},
                    {},
                    {
                        'cpu': {
                            'stepping': {
                                '_op': '!=',
                                '_value': '10',
                            },
                        },
                    },
                    {
                        'cpu': {
                            'speed': {
                                '_op': '>=',
                                '_value': '2300.0',
                            },
                        },
                    },
                    {},
                    {
                        'not': {'cpu': {'model_name': {'_op': 'like', '_value': 'Haswell'}}},
                    },
                    {
                        'cpu': {
                            'vendor': {
                                '_op': 'like',
                                '_value': 'Intel%',
                            },
                        },
                    },
                    {
                        'and': [
                            {'cpu': {'flag': {'_op': '==', '_value': 'avx'}}},
                            {'cpu': {'flag': {'_op': '==', '_value': 'avx2'}}},
                            {'cpu': {'flag': {'_op': '!=', '_value': 'smep'}}},
                        ]
                    },
                    {'cpu': {'hyper': {'_op': '==', '_value': 'True'}}},
                ]
            },
            {'and': [{}, {}, {}, {}, {}]},
            {'and': [{}, {}, {}, {}, {}]},
            {'system': {'memory': {'_op': '==', '_value': '8192'}}},
            {
                'and': [
                    {
                        'and': [
                            {'disk': {'size': {'_op': '==', '_value': '42949672960'}}},
                            {'disk': {'phys_sector_size': {'_op': '==', '_value': '4096 B'}}},
                            {'disk': {'model': {'_op': 'like', '_value': 'WD 100G%'}}},
                        ]
                    },
                    {
                        'and': [
                            {'disk': {'size': {'_op': '==', '_value': '128849018880'}}},
                            {'disk': {'sector_size': {'_op': '==', '_value': '512 B'}}},
                            {'key_value': {'_key': 'BOOTDISK', '_op': '==', '_value': 'virtblk'}},
                        ]
                    },
                ]
            },
            {
                'and': [
                    {
                        'and': [
                            {},
                            {},
                            {},
                            {},
                            {},
                            {},
                        ]
                    },
                    {},
                ],
            },
            {'hostname': {'_op': 'like', '_value': '%.foo.redhat.com'}},
            {'labcontroller': {'_op': '!=', '_value': 'lab-1.bar.redhat.com'}},
            {
                'and': [
                    {},
                    {
                        'system': {
                            'vendor': {
                                '_op': 'like',
                                '_value': 'Dell%',
                            },
                        },
                    },
                    {},
                    {
                        'system': {
                            'numanodes': {
                                '_op': '<',
                                '_value': '4',
                            },
                        },
                    },
                    {
                        'system': {
                            'model': {
                                '_op': 'like',
                                '_value': 'PowerEdge R750',
                            },
                        },
                    },
                ]
            },
            {
                'key_value': {
                    '_key': 'TPM',
                    '_op': '==',
                    '_value': '2.0',
                },
            },
            {
                'and': [
                    {'system': {'hypervisor': {'_op': '==', '_value': ''}}},
                    {},
                    {'system': {'hypervisor': {'_op': 'like', '_value': 'xen'}}},
                ]
            },
            {
                'and': [
                    {
                        'system': {
                            'key_value': {'_key': 'ZCRYPT_MODEL', '_op': '==', '_value': 'CEX8C'}
                        }
                    },
                    {
                        'system': {
                            'key_value': {'_key': 'ZCRYPT_MODE', '_op': '==', '_value': 'CCA'}
                        }
                    },
                ]
            },
        ]
    }


def test_cpu_model(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.model'](_parse_cpu({'model': '79'}), root_logger)

    assert result.to_mrack() == {'cpu': {'model': {'_op': '==', '_value': '79'}}}


def test_cpu_processors(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.processors'](
        _parse_cpu({'processors': '79'}), root_logger
    )

    assert result.to_mrack() == {'cpu': {'processors': {'_op': '==', '_value': '79'}}}


def test_cpu_cores(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.cores'](_parse_cpu({'cores': '2'}), root_logger)

    assert result.to_mrack() == {'cpu': {'cores': {'_op': '==', '_value': '2'}}}


def test_cpu_stepping(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.stepping'](_parse_cpu({'stepping': '10'}), root_logger)

    assert result.to_mrack() == {'cpu': {'stepping': {'_op': '==', '_value': '10'}}}


def test_cpu_vendor_name(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': 'GenuineIntel'}), root_logger
    )

    assert result.to_mrack() == {'cpu': {'vendor': {'_op': '==', '_value': 'GenuineIntel'}}}

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '!= GenuineIntel'}), root_logger
    )

    assert result.to_mrack() == {'cpu': {'vendor': {'_op': '!=', '_value': 'GenuineIntel'}}}

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '~ .*Intel'}), root_logger
    )

    assert result.to_mrack() == {'cpu': {'vendor': {'_op': 'like', '_value': '%Intel'}}}

    result = _CONSTRAINT_TRANSFORMERS['cpu.vendor_name'](
        _parse_cpu({'vendor-name': '!~ .*Intel'}), root_logger
    )

    assert result.to_mrack() == {'not': {'cpu': {'vendor': {'_op': 'like', '_value': '%Intel'}}}}


def test_disk_driver(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': 'mpt3sas'}, 1), root_logger
    )

    assert result.to_mrack() == {
        'key_value': {'_key': 'BOOTDISK', '_op': '==', '_value': 'mpt3sas'}
    }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '!= mpt3sas'}, 1), root_logger
    )

    assert result.to_mrack() == {
        'key_value': {'_key': 'BOOTDISK', '_op': '!=', '_value': 'mpt3sas'}
    }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '~ mpt3.*'}, 1), root_logger
    )

    assert result.to_mrack() == {
        'key_value': {'_key': 'BOOTDISK', '_op': 'like', '_value': 'mpt3%'}
    }

    result = _CONSTRAINT_TRANSFORMERS['disk.driver'](
        _parse_disk({'driver': '!~ mpt3.*'}, 1), root_logger
    )

    assert result.to_mrack() == {
        'not': {'key_value': {'_key': 'BOOTDISK', '_op': 'like', '_value': 'mpt3%'}}
    }


def test_disk_size(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.size'](
        _parse_disk({'size': '>= 40 GiB'}, 1), root_logger
    )

    assert result.to_mrack() == {'disk': {'size': {'_op': '>=', '_value': '42949672960'}}}


def test_disk_model_name(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': 'PERC H310'}, 1), root_logger
    )

    assert result.to_mrack() == {'disk': {'model': {'_op': '==', '_value': 'PERC H310'}}}

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '!= PERC H310'}, 1), root_logger
    )

    assert result.to_mrack() == {'disk': {'model': {'_op': '!=', '_value': 'PERC H310'}}}

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '~ PERC.*'}, 1), root_logger
    )

    assert result.to_mrack() == {'disk': {'model': {'_op': 'like', '_value': 'PERC%'}}}

    result = _CONSTRAINT_TRANSFORMERS['disk.model_name'](
        _parse_disk({'model-name': '!~ PERC.*'}, 1), root_logger
    )

    assert result.to_mrack() == {'not': {'disk': {'model': {'_op': 'like', '_value': 'PERC%'}}}}


def test_memory(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['memory'](_parse_memory({'memory': '>= 4 GiB'}), root_logger)

    assert result.to_mrack() == {'system': {'memory': {'_op': '>=', '_value': '4096'}}}


def test_hostname(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': 'foo.dot.com'}), root_logger
    )

    assert result.to_mrack() == {'hostname': {'_op': '==', '_value': 'foo.dot.com'}}

    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': '~ foo.*.dot.+.com'}), root_logger
    )

    assert result.to_mrack() == {'hostname': {'_op': 'like', '_value': 'foo%.dot%.com'}}

    result = _CONSTRAINT_TRANSFORMERS['hostname'](
        _parse_hostname({'hostname': '!~ foo.*.dot.+.com'}), root_logger
    )

    assert result.to_mrack() == {'not': {'hostname': {'_op': 'like', '_value': 'foo%.dot%.com'}}}


def test_virtualization_is_virtualized(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['virtualization.is_virtualized'](
        _parse_virtualization({'is-virtualized': True}), root_logger
    )

    assert result.to_mrack() == {'system': {'hypervisor': {'_op': '!=', '_value': ''}}}

    result = _CONSTRAINT_TRANSFORMERS['virtualization.is_virtualized'](
        _parse_virtualization({'is-virtualized': False}), root_logger
    )

    assert result.to_mrack() == {'system': {'hypervisor': {'_op': '==', '_value': ''}}}


def test_virtualization_hypervisor(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "~ kvm"}), root_logger
    )

    assert result.to_mrack() == {'system': {'hypervisor': {'_op': 'like', '_value': 'kvm'}}}

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "!~ kvm"}), root_logger
    )

    assert result.to_mrack() == {
        'not': {'system': {'hypervisor': {'_op': 'like', '_value': 'kvm'}}}
    }

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "kvm"}), root_logger
    )

    assert result.to_mrack() == {'system': {'hypervisor': {'_op': '==', '_value': 'kvm'}}}

    result = _CONSTRAINT_TRANSFORMERS['virtualization.hypervisor'](
        _parse_virtualization({"hypervisor": "!= kvm"}), root_logger
    )

    assert result.to_mrack() == {'system': {'hypervisor': {'_op': '!=', '_value': 'kvm'}}}


def test_zcrypt_adapter(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': 'CEX8C'}), root_logger
    )

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODEL', '_op': '==', '_value': 'CEX8C'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '!= CEX8C'}), root_logger
    )

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODEL', '_op': '!=', '_value': 'CEX8C'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '~ CEX.*'}), root_logger
    )

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODEL', '_op': 'like', '_value': 'CEX%'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.adapter'](
        _parse_zcrypt({'adapter': '!~ CEX.*'}), root_logger
    )

    assert result.to_mrack() == {
        'not': {'system': {'key_value': {'_key': 'ZCRYPT_MODEL', '_op': 'like', '_value': 'CEX%'}}}
    }


def test_zcrypt_mode(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](_parse_zcrypt({'mode': 'CCA'}), root_logger)

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODE', '_op': '==', '_value': 'CCA'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '!= CCA'}), root_logger
    )

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODE', '_op': '!=', '_value': 'CCA'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '~ C.*A'}), root_logger
    )

    assert result.to_mrack() == {
        'system': {'key_value': {'_key': 'ZCRYPT_MODE', '_op': 'like', '_value': 'C%A'}}
    }

    result = _CONSTRAINT_TRANSFORMERS['zcrypt.mode'](
        _parse_zcrypt({'mode': '!~ C.*A'}), root_logger
    )

    assert result.to_mrack() == {
        'not': {'system': {'key_value': {'_key': 'ZCRYPT_MODE', '_op': 'like', '_value': 'C%A'}}}
    }


def test_iommu_is_supported(root_logger: Logger) -> None:
    for value in True, False:
        result = _CONSTRAINT_TRANSFORMERS['iommu.is_supported'](
            _parse_iommu({"is-supported": value}), root_logger
        )

        assert result.to_mrack() == {
            'key_value': {'_key': 'VIRT_IOMMU', '_op': '==', '_value': str(int(value))}
        }


def test_location_lab_controller(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['location.lab_controller'](
        _parse_location({"lab-controller": "lab-01.bar.redhat.com"}), root_logger
    )

    assert result.to_mrack() == {'labcontroller': {'_op': '==', '_value': 'lab-01.bar.redhat.com'}}

    result = _CONSTRAINT_TRANSFORMERS['location.lab_controller'](
        _parse_location({"lab-controller": "!= lab-01.bar.redhat.com"}), root_logger
    )

    assert result.to_mrack() == {'labcontroller': {'_op': '!=', '_value': 'lab-01.bar.redhat.com'}}


def test_tpm_version(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['tpm.version'](_parse_tpm({'version': '2.0'}), root_logger)

    assert result.to_mrack() == {'key_value': {'_key': 'TPM', '_op': '==', '_value': '2.0'}}

    result = _CONSTRAINT_TRANSFORMERS['tpm.version'](
        _parse_tpm({'version': '!= 2.0'}), root_logger
    )

    assert result.to_mrack() == {'key_value': {'_key': 'TPM', '_op': '!=', '_value': '2.0'}}


def test_system_numa_nodes(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['system.numa_nodes'](
        _parse_system({'numa-nodes': '2'}), root_logger
    )

    assert result.to_mrack() == {'system': {'numanodes': {'_op': '==', '_value': '2'}}}


def test_system_model_name(root_logger: Logger) -> None:
    result = _CONSTRAINT_TRANSFORMERS['system.model_name'](
        _parse_system({'model-name': '!~ PowerEdge R750.*'}), root_logger
    )

    assert result.to_mrack() == {
        'not': {'system': {'model': {'_op': 'like', '_value': 'PowerEdge R750%'}}}
    }


def test_or_constraint(root_logger: Logger) -> None:
    hw = Hardware.from_spec(tmt.utils.yaml_to_dict(textwrap.dedent(OR_HARDWARE_REQUIREMENTS)))
    assert hw.constraint is not None

    result = constraint_to_beaker_filter(hw.constraint, root_logger)
    assert result.to_mrack() == {
        'or': [
            {
                'hostname': {
                    '_op': '==',
                    '_value': 'dummy1.redhat.com',
                },
            },
            {
                'or': [
                    {
                        'hostname': {
                            '_op': '==',
                            '_value': 'dummy2.redhat.com',
                        },
                    },
                    {
                        'or': [
                            {
                                'hostname': {
                                    '_op': '==',
                                    '_value': 'dummy3.redhat.com',
                                },
                            },
                            {
                                'hostname': {
                                    '_op': '==',
                                    '_value': 'dummy4.redhat.com',
                                },
                            },
                        ],
                    },
                ],
            },
        ],
    }
